from apscheduler.triggers.combining import AndTrigger, OrTrigger
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from flask import Blueprint, request, jsonify
from flask_apscheduler import api
from flask_security import roles_accepted
from opentakserver.extensions import apscheduler

scheduler_api_blueprint = Blueprint('schedule_api_blueprint', __name__)


@scheduler_api_blueprint.route("/api/scheduler/", strict_slashes=False)
@roles_accepted("administrator")
def scheduler_info():
    return api.get_scheduler_info()


@scheduler_api_blueprint.route("/api/scheduler/jobs", strict_slashes=False)
@roles_accepted("administrator")
def get_jobs():
    return api.get_jobs()


@scheduler_api_blueprint.route("/api/scheduler/job/pause", methods=['POST'], strict_slashes=False)
@roles_accepted("administrator")
def pause_job():
    if 'job_id' not in request.json:
        return {'success': False, 'errors': 'Please provide a job_id'}, 400, {'Content-Type': 'application/json'}
    try:
        return api.pause_job(request.json['job_id'])
    except BaseException as e:
        return {'success': False, 'error': str(e)}, 400, {'Content-Type': 'application/json'}


@scheduler_api_blueprint.route("/api/scheduler/job/resume", methods=['POST'], strict_slashes=False)
@roles_accepted("administrator")
def resume_job():
    if 'job_id' not in request.json:
        return {'success': False, 'errors': 'Please provide a job_id'}, 400, {'Content-Type': 'application/json'}
    try:
        #  0 == STATE_STOPPED
        if apscheduler.state == 0:
            apscheduler.start()
        #  2 == STATE_PAUSED
        elif apscheduler.state == 2:
            apscheduler.resume()

        return api.resume_job(request.json['job_id'])
    except BaseException as e:
        return {'success': False, 'error': str(e)}, 400, {'Content-Type': 'application/json'}


@scheduler_api_blueprint.route("/api/scheduler/job/run", methods=['POST'], strict_slashes=False)
@roles_accepted("administrator")
def run_job():
    if 'job_id' not in request.json:
        return {'success': False, 'errors': 'Please provide a job_id'}, 400, {'Content-Type': 'application/json'}
    try:
        return api.run_job(request.json['job_id'])
    except BaseException as e:
        return {'success': False, 'error': str(e)}, 400, {'Content-Type': 'application/json'}


@scheduler_api_blueprint.route("/api/scheduler/job/modify", methods=['POST'], strict_slashes=False)
@roles_accepted("administrator")
def modify_job():
    if 'job_id' not in request.json:
        return {'success': False, 'errors': 'Please provide a job_id'}, 400, {'Content-Type': 'application/json'}
    try:
        job_id = request.json.pop('job_id')
        trigger = request.json.pop('trigger')
        job = apscheduler.get_job(job_id)
        if trigger == 'interval':
            trigger = IntervalTrigger(**request.json)
        elif trigger == 'cron':
            trigger = CronTrigger(**request.json)
        elif trigger == 'date':
            trigger = DateTrigger(**request.json)
        elif trigger == 'and':
            trigger = AndTrigger(**request.json)
        elif trigger == 'or':
            trigger = OrTrigger(**request.json)
        else:
            return jsonify({'success': False, 'error': 'Invalid trigger: {}'.format(trigger)}), 400

        job.modify(trigger=trigger)
        return jsonify({'success': True})
    except BaseException as e:
        return {'success': False, 'error': str(e)}, 400, {'Content-Type': 'application/json'}



import functools

from flask import request, Blueprint
from flask_security import current_user
from flask_socketio import disconnect

from opentakserver.extensions import logger, socketio

ots_socketio_blueprint = Blueprint('ots_socketio_blueprint', __name__)


def authenticated_only(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if not current_user.is_authenticated:
            logger.debug("Disconnecting {} from {}".format(request.sid, request.namespace))
            disconnect(request.sid, namespace=request.namespace)
        else:
            return f(*args, **kwargs)

    return wrapped


@socketio.on('connect', namespace="/socket.io")
@authenticated_only
def connect(data):
    logger.info('got a socketio connection from {}'.format(current_user.username))


@socketio.on('message', namespace="/socket.io")
@authenticated_only
def message(message):
    logger.info("Got a message".format(message))



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import String
from flask_security.models import fsqla_v3 as fsqla
from sqlalchemy.orm import relationship


@dataclass
class User(db.Model, fsqla.FsUserMixin):
    email = db.Column(String, nullable=True)
    video_streams = relationship("VideoStream", back_populates="user")
    euds = relationship("EUD", back_populates="user")
    data_packages = relationship("DataPackage", back_populates="user")
    certificate = relationship("Certificate", back_populates="user")

    def serialize(self):
        return {
            'username': self.username,
            'active': self.active,
            'last_login_at': self.last_login_at,
            'last_login_ip': self.last_login_ip,
            'current_login_at': self.current_login_at,
            'current_login_ip': self.current_login_ip,
            'email': self.email,
            'login_count': self.login_count,
            'euds': [eud.serialize() for eud in self.euds],
            'video_streams': [v.serialize() for v in self.video_streams],
            'roles': [role.serialize() for role in self.roles]
        }

    def to_json(self):
        return self.serialize()



import json
import socket
import traceback
import uuid
from xml.etree.ElementTree import Element, SubElement, tostring, fromstring, ParseError
import datetime
from threading import Thread

from flask_security import verify_password

from bs4 import BeautifulSoup
import pika

from opentakserver.extensions import db
from opentakserver.models.EUD import EUD


class ClientController(Thread):
    def __init__(self, address, port, sock, logger, app, is_ssl):
        Thread.__init__(self)
        self.address = address
        self.port = port
        self.sock = sock
        self.logger = logger
        self.shutdown = False
        self.sock.settimeout(1.0)
        self.app = app
        self.db = db
        self.is_ssl = is_ssl

        # Device attributes
        self.uid = None
        self.device = None
        self.os = None
        self.platform = None
        self.version = None
        self.callsign = None
        self.phone_number = None
        self.battery = None
        self.groups = {}
        self.device_inserted = False
        self.is_authenticated = False

        # Location Attributes
        self.latitude = 0
        self.longitude = 0
        self.ce = 0
        self.hae = 0
        self.le = 0
        self.course = 0
        self.speed = 0
        self.location_source = None
        self.common_name = None

        if self.is_ssl:
            try:
                self.sock.do_handshake()
                for c in self.sock.getpeercert()['subject']:
                    if c[0][0] == 'commonName':
                        self.common_name = c[0][1]
                        self.logger.debug("Got common name {}".format(self.common_name))
            except BaseException as e:
                logger.warning("Failed to do handshake: {}".format(e))

        # RabbitMQ
        try:
            self.rabbit_connection = pika.SelectConnection(pika.ConnectionParameters(self.app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")),
                                                           self.on_connection_open)
            self.rabbit_channel = None
            self.iothread = Thread(target=self.rabbit_connection.ioloop.start)
            self.iothread.daemon = True
            self.iothread.start()
            self.is_consuming = False
        except BaseException as e:
            self.logger.error("Failed to connect to rabbitmq: {}".format(e))
            return

    def on_connection_open(self, connection):
        self.rabbit_connection.channel(on_open_callback=self.on_channel_open)
        self.rabbit_connection.add_on_close_callback(self.on_close)

    def on_channel_open(self, channel):
        self.rabbit_channel = channel
        self.rabbit_channel.add_on_close_callback(self.on_close)

    def on_close(self, channel, error):
        self.logger.info("Connection closed for {}: {}".format(self.address, error))

    def on_message(self, unused_channel, basic_deliver, properties, body):
        try:
            body = json.loads(body)
            if body['uid'] != self.uid:
                self.sock.send(body['cot'].encode())
        except:
            self.logger.error(traceback.format_exc())

    def run(self):
        while not self.shutdown:
            try:
                data = self.sock.recv(4096)
            except (ConnectionError, ConnectionResetError) as e:
                self.send_disconnect_cot()
                break
            except TimeoutError:
                if self.shutdown:
                    self.logger.warning("Closing connection to {}".format(self.address))
                    self.send_disconnect_cot()
                    self.sock.shutdown(socket.SHUT_RDWR)
                    self.sock.close()
                    break
                else:
                    continue

            if data:
                # Sometimes recv() doesn't get all of the XML data in one go. Test if the XML is well-formed
                # and if not, call recv() until it is
                while True:
                    try:
                        fromstring(data)
                        break
                    except ParseError as e:
                        try:
                            data += self.sock.recv(4096)
                            break
                        except (ConnectionError, TimeoutError, ConnectionResetError) as e:
                            break

                soup = BeautifulSoup(data, 'xml')

                event = soup.find('event')
                auth = soup.find('auth')

                if not self.is_authenticated and (auth or self.common_name):
                    with self.app.app_context():
                        if auth:
                            cot = auth.find('cot')
                            if cot:
                                username = cot.attrs['username']
                                password = cot.attrs['password']
                                uid = cot.attrs['uid']
                                user = self.app.security.datastore.find_user(username=username)
                        elif self.common_name:
                            user = self.app.security.datastore.find_user(username=self.common_name)

                        if not user:
                            self.logger.warning("User {} does not exist".format(username))
                            self.close_connection()
                            break
                        elif not user.active:
                            self.logger.warning("User {} is deactivated, disconnecting".format(username))
                            self.close_connection()
                            break
                        elif self.common_name:
                            self.logger.info("{} is ID'ed by cert".format(user.username))
                            self.is_authenticated = True
                        elif verify_password(password, user.password):
                            self.logger.info("Successful login from {}".format(username))
                            self.is_authenticated = True
                            try:
                                eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=uid)).first()[0]
                                self.logger.debug("Associating EUD uid {} to user {}".format(eud.uid, user.username))
                                eud.user_id = user.id
                                self.db.session.commit()
                            except:
                                self.logger.debug("This is a new eud: {} {}".format(uid, user.username))
                                eud = EUD()
                                eud.uid = uid
                                eud.user_id = user.id
                                self.db.session.add(eud)
                                self.db.session.commit()

                            else:
                                self.logger.warning("Wrong password for user {}".format(username))
                                self.close_connection()
                                break

                if event:
                    # If this client is connected via ssl, make sure they're authenticated
                    # before accepting any data from them
                    if self.is_ssl and not self.is_authenticated:
                        self.logger.warning("EUD isn't authenticated, ignoring")
                        continue

                    if event and self.pong(event):
                        continue

                    if event and not self.uid:
                        self.parse_device_info(event)

                    message = {'uid': self.uid, 'cot': str(soup)}
                    if self.rabbit_channel:
                        self.rabbit_channel.basic_publish(exchange='cot_controller', routing_key='',
                                                          body=json.dumps(message))

            else:
                self.send_disconnect_cot()
                break

    def close_connection(self):
        self.send_disconnect_cot()
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def stop(self):
        self.shutdown = True

    def pong(self, event):
        if 'uid' in event.attrs and event.attrs['uid'].endswith('ping'):
            now = datetime.datetime.now()
            stale = now + datetime.timedelta(seconds=10)

            cot = Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-c-t-r', 'version': '2.0',
                                      'uid': "{}-pong".format(self.uid), 'start': now, 'time': now, 'stale': stale})
            SubElement(cot, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                                'lon': '0'})

            self.sock.send(event.encode())
            return True

        return False

    def parse_device_info(self, event):
        if not self.rabbit_channel:
            return
        self.uid = event.attrs['uid']
        contact = event.find('contact')
        if contact:
            if 'callsign' in contact.attrs:
                self.callsign = contact.attrs['callsign']

                self.logger.debug("Declaring queue {}".format(self.uid))
                self.rabbit_channel.queue_declare(queue=self.uid)
                self.rabbit_channel.queue_bind(exchange='cot', queue=self.uid)
                self.rabbit_channel.basic_consume(queue=self.uid, on_message_callback=self.on_message, auto_ack=True)
                self.logger.debug("{} is consuming".format(self.callsign))

    def send_disconnect_cot(self):
        if self.uid:
            now = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
            stale = (datetime.datetime.now() + datetime.timedelta(seconds=10)).strftime("%Y-%m-%dT%H:%M:%SZ")

            event = Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-d-d', 'version': '2.0',
                                      'uid': str(uuid.uuid4()), 'start': now, 'time': now, 'stale': stale})
            point = SubElement(event, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                                'lon': '0'})
            detail = SubElement(event, 'detail')
            link = SubElement(detail, 'link', {'relation': 'p-p', 'uid': self.uid, 'type': 'a-f-G-U-C'})
            flow_tags = SubElement(detail, '_flow-tags_', {'TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0': now})

            message = json.dumps({'uid': self.uid, 'cot': tostring(event).decode('utf-8')})
            self.rabbit_channel.basic_publish(exchange='cot_controller', routing_key='', body=message)
        self.logger.info('{} disconnected'.format(self.address))
        if self.rabbit_connection:
            self.rabbit_connection.close()



from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Float, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


class Point(db.Model):
    __tablename__ = "points"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String)
    device_uid: Mapped[str] = mapped_column(String, ForeignKey("euds.uid"))
    latitude: Mapped[float] = mapped_column(Float, nullable=True)
    longitude: Mapped[float] = mapped_column(Float, nullable=True)
    ce: Mapped[float] = mapped_column(Float, nullable=True)
    hae: Mapped[float] = mapped_column(Float, nullable=True)
    le: Mapped[float] = mapped_column(Float, nullable=True)
    course: Mapped[float] = mapped_column(Float, nullable=True)
    speed: Mapped[float] = mapped_column(Float, nullable=True)
    location_source: Mapped[str] = mapped_column(String, nullable=True)
    battery: Mapped[float] = mapped_column(Float, nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)
    azimuth: Mapped[float] = mapped_column(Float, nullable=True)
    # Camera field of view from TAK ICU and OpenTAK ICU
    fov: Mapped[float] = mapped_column(Float, nullable=True)
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id"), nullable=True)
    cot = relationship("CoT", back_populates="point")

    # Only populate this field of the CoT type matches ^a- and how matches either ^m-g or ^h-e
    eud = relationship("EUD", back_populates="points")
    casevac = relationship("CasEvac", back_populates="point")
    geochat = relationship("GeoChat", back_populates="point")
    alert = relationship("Alert", back_populates="point")
    marker: Mapped["Marker"] = relationship(back_populates="point")
    rb_line = relationship("RBLine", back_populates="point")

    def serialize(self):
        return {
            'uid': self.uid,
            'device_uid': self.device_uid,
            'latitude': self.latitude,
            'longitude': self.longitude,
            'ce': self.ce,
            'hae': self.hae,
            'le': self.le,
            'course': self.course,
            'speed': self.speed,
            'azimuth': self.azimuth,
            'fov': self.fov,
            'location_source': self.location_source,
            'battery': self.battery,
            'timestamp': self.timestamp,
        }

    def to_json(self):
        return {
            'uid': self.uid,
            'device_uid': self.device_uid,
            'latitude': self.latitude,
            'longitude': self.longitude,
            'ce': self.ce,
            'hae': self.hae,
            'le': self.le,
            'course': self.course,
            'speed': self.speed,
            'azimuth': self.azimuth,
            'fov': self.fov,
            'location_source': self.location_source,
            'battery': self.battery,
            'timestamp': iso8601_string_from_datetime(self.timestamp),
            'how': self.cot.how,
            'type': self.cot.type,
            'callsign': self.eud.callsign if self.eud else None
        }



import colorlog
from flask_sqlalchemy import SQLAlchemy
from flask_socketio import SocketIO
from opentakserver.models.Base import Base
from flask_mailman import Mail
from flask_apscheduler import APScheduler

logger = colorlog.getLogger('OpenTAKServer')

mail = Mail()

apscheduler = APScheduler()

db = SQLAlchemy(model_class=Base)

socketio = SocketIO(async_mode='eventlet')



import json
import traceback

import pika
from bs4 import BeautifulSoup
from flask import Blueprint
import adsbxcot
from flask import current_app as app

from opentakserver.extensions import apscheduler, logger, db
import requests

from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.Alert import Alert
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.Certificate import Certificate
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.CoT import CoT
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.EUD import EUD
from opentakserver.models.GeoChat import GeoChat
from opentakserver.models.Marker import Marker
from opentakserver.models.Point import Point
from opentakserver.models.RBLine import RBLine
from opentakserver.models.Team import Team
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.ZMIST import ZMIST

scheduler_blueprint = Blueprint('scheduler_blueprint', __name__)


@apscheduler.task("interval", name="Airplanes.live", id='get_airplanes_live_data', next_run_time=None)
def get_airplanes_live_data():
    with apscheduler.app.app_context():
        try:
            r = requests.get('https://api.airplanes.live/v2/point/{}/{}/{}'
                             .format(app.config["OTS_AIRPLANES_LIVE_LAT"],
                                     app.config["OTS_AIRPLANES_LIVE_LON"],
                                     app.config["OTS_AIRPLANES_LIVE_RADIUS"]))
            if r.status_code == 200:
                rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")))
                channel = rabbit_connection.channel()

                for craft in r.json()['ac']:
                    try:
                        event = adsbxcot.adsbx_to_cot(craft, known_craft=None)
                    except ValueError:
                        continue

                    # noinspection PyTypeChecker
                    channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
                        {'cot': str(BeautifulSoup(event, 'xml')), 'uid': app.config['OTS_NODE_ID']}))
                    # noinspection PyTypeChecker
                    channel.basic_publish(exchange='cot_controller', routing_key='', body=json.dumps(
                        {'cot': str(BeautifulSoup(event, 'xml')), 'uid': app.config['OTS_NODE_ID']}))

                channel.close()
                rabbit_connection.close()
            else:
                logger.error('Failed to get airplanes.live: {}'.format(r.text))
        except BaseException as e:
            logger.error("Failed to get airplanes.live: {}".format(e))
            logger.error(traceback.format_exc())


@apscheduler.task("interval", name="Delete Video Recordings", id="delete_video_recordings", next_run_time=None)
def delete_video_recordings():
    with apscheduler.app.app_context():
        VideoRecording.query.delete()
        db.session.commit()

        try:
            r = requests.get('{}/v3/recordings/list'.format(app.config.get("OTS_MEDIAMTX_API_ADDRESS")))
            if r.status_code == 200:
                recordings = r.json()
                for path in recordings['items']:
                    for recording in path['segments']:
                        r = requests.delete('{}/v3/recordings/deletesegment'.format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"),),
                                            params={'start': recording['start'], 'path': path['name']})
                        if r.status_code != 200:
                            logger.error(
                                "Failed to delete {} from {}: {}".format(recording['start'], path['name'], r.text))
        except BaseException as e:
            logger.error("Failed to delete recordings: {}".format(e))


@apscheduler.task("cron", id="purge_data", name="Purge Data", day="*", hour=0, minute=0, next_run_time=None)
def purge_data():
    delete_video_recordings()
    Alert.query.delete()
    CasEvac.query.delete()
    Certificate.query.delete()
    Chatroom.query.delete()
    ChatroomsUids.query.delete()
    CoT.query.delete()
    DataPackage.query.delete()
    EUD.query.delete()
    GeoChat.query.delete()
    Marker.query.delete()
    Point.query.delete()
    RBLine.query.delete()
    Team.query.delete()
    VideoStream.query.delete()
    ZMIST.query.delete()
    db.session.commit()
    logger.info("Purged all data")



import sys
import traceback

import eventlet
try:
    eventlet.monkey_patch()
except BaseException as e:
    print("Failed to monkey_patch(): {}".format(e))

from opentakserver.PasswordValidator import PasswordValidator

import platform
import requests
from sqlalchemy import insert
import sqlite3
from opentakserver.models.Icon import Icon

import yaml
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore

from opentakserver.EmailValidator import EmailValidator

import logging
from logging.handlers import TimedRotatingFileHandler
import os

import colorlog
from werkzeug.middleware.proxy_fix import ProxyFix
from datetime import datetime
import sqlalchemy

import flask_wtf

import pika
from flask import Flask, jsonify
from flask_cors import CORS

from flask_security import Security, SQLAlchemyUserDatastore, hash_password, uia_username_mapper, uia_email_mapper
from flask_security.models import fsqla_v3 as fsqla
from flask_security.signals import user_registered

from opentakserver.extensions import logger, db, socketio, mail, apscheduler
from opentakserver.defaultconfig import DefaultConfig
from opentakserver.models.WebAuthn import WebAuthn

from opentakserver.controllers.cot_controller import CoTController
from opentakserver.certificate_authority import CertificateAuthority
from opentakserver.SocketServer import SocketServer
try:
    from opentakserver.mumble.mumble_ice_app import MumbleIceDaemon
except ModuleNotFoundError:
    print("Mumble auth not supported on this platform")


def init_extensions(app):
    db.init_app(app)

    # Handle config options that can't be serialized to yaml
    app.config.update({"SCHEDULER_JOBSTORES": {'default': SQLAlchemyJobStore(url=app.config.get("SQLALCHEMY_DATABASE_URI"))}})
    identity_attributes = [{"username": {"mapper": uia_username_mapper, "case_insensitive": False}}]

    # Don't allow registration unless email is enabled
    if app.config.get("OTS_ENABLE_EMAIL"):
        identity_attributes.append({"email": {"mapper": uia_email_mapper, "case_insensitive": True}})
        app.config.update({
            "SECURITY_REGISTERABLE": True,
            "SECURITY_CONFIRMABLE": True,
            "SECURITY_RECOVERABLE": True,
            "SECURITY_TWO_FACTOR_ENABLED_METHODS": ["authenticator", "email"]
        })
    else:
        app.config.update({
            "SECURITY_REGISTERABLE": False,
            "SECURITY_CONFIRMABLE": False,
            "SECURITY_RECOVERABLE": False,
            "SECURITY_TWO_FACTOR_ENABLED_METHODS": ["authenticator"]
        })
    app.config.update({"SECURITY_USER_IDENTITY_ATTRIBUTES": identity_attributes})

    ca = CertificateAuthority(logger, app)
    ca.create_ca()

    cors = CORS(app, resources={r"/api/*": {"origins": "*"}, r"/Marti/*": {"origins": "*"}, r"/*": {"origins": "*"}},
                supports_credentials=True)
    flask_wtf.CSRFProtect(app)

    socketio_logger = False
    if app.config.get("DEBUG"):
        socketio_logger = logger
    socketio.init_app(app, logger=socketio_logger)

    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")))
    channel = rabbit_connection.channel()
    channel.exchange_declare('cot', durable=True, exchange_type='fanout')
    channel.exchange_declare('dms', durable=True, exchange_type='direct')
    channel.exchange_declare('chatrooms', durable=True, exchange_type='direct')
    channel.queue_declare(queue='cot_controller')
    channel.exchange_declare(exchange='cot_controller', exchange_type='fanout')

    cot_thread = CoTController(app.app_context(), logger, db, socketio)
    app.cot_thread = cot_thread

    if not apscheduler.running:
        apscheduler.init_app(app)
        apscheduler.start(paused=True)

    try:
        fsqla.FsModels.set_db_info(db)
    except sqlalchemy.exc.InvalidRequestError:
        pass

    from opentakserver.models.user import User
    from opentakserver.models.role import Role

    user_datastore = SQLAlchemyUserDatastore(db, User, Role, WebAuthn)
    app.security = Security(app, user_datastore, mail_util_cls=EmailValidator, password_util_cls=PasswordValidator)

    mail.init_app(app)
    with app.app_context():
        db.create_all()


def setup_logging(app):
    level = logging.INFO
    if app.config.get("DEBUG"):
        level = logging.DEBUG

    if sys.stdout.isatty():
        color_log_handler = colorlog.StreamHandler()
        color_log_formatter = colorlog.ColoredFormatter(
            '%(log_color)s[%(asctime)s] - OpenTAKServer[%(process)d] - %(module)s - %(levelname)s - %(message)s', datefmt="%Y-%m-%d %H:%M:%S")
        color_log_handler.setFormatter(color_log_formatter)
        logger.setLevel(level)
        logger.addHandler(color_log_handler)
        logger.info("Added color logger")

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs"), exist_ok=True)
    fh = TimedRotatingFileHandler(os.path.join(app.config.get("OTS_DATA_FOLDER"), 'logs', 'opentakserver.log'), when='midnight', backupCount=app.config.get("OTS_BACKUP_COUNT"))
    fh.setLevel(level)
    fh.setFormatter(logging.Formatter("[%(asctime)s] - OpenTAKServer[%(process)d] - %(module)s - %(levelname)s - %(message)s"))
    logger.addHandler(fh)


def create_app():
    app = Flask(__name__)
    app.config.from_object(DefaultConfig)
    setup_logging(app)

    # Load config.yml if it exists
    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
        app.config.from_file(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load)
    else:
        # First run, created config.yml based on default settings
        logger.info("Creating config.yml")
        with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
            conf = {}
            for option in DefaultConfig.__dict__:
                # Fix the sqlite DB path on Windows
                if option == "SQLALCHEMY_DATABASE_URI" and platform.system() == "Windows" and DefaultConfig.__dict__[option].startswith("sqlite"):
                    conf[option] = DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
                elif option.isupper():
                    conf[option] = DefaultConfig.__dict__[option]
            config.write(yaml.safe_dump(conf))

        # Try to set the MediaMTX token
        try:
            with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "mediamtx.yml"), "r") as mediamtx_config:
                conf = mediamtx_config.read()
                conf = conf.replace("MTX_TOKEN", app.config.get("OTS_MEDIAMTX_TOKEN"))
            with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "mediamtx.yml"), "w") as mediamtx_config:
                mediamtx_config.write(conf)
        except BaseException as e:
            logger.error("Failed to set MediaMTX token: {}".format(e))

    init_extensions(app)

    from opentakserver.blueprints.marti import marti_blueprint
    app.register_blueprint(marti_blueprint)

    from opentakserver.blueprints.api import api_blueprint
    app.register_blueprint(api_blueprint)

    from opentakserver.blueprints.ots_socketio import ots_socketio_blueprint
    app.register_blueprint(ots_socketio_blueprint)

    from opentakserver.blueprints.scheduled_jobs import scheduler_blueprint
    app.register_blueprint(scheduler_blueprint)

    from opentakserver.blueprints.scheduler_api import scheduler_api_blueprint
    app.register_blueprint(scheduler_api_blueprint)

    from opentakserver.blueprints.config import config_blueprint
    app.register_blueprint(config_blueprint)

    app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_host=1)

    return app


app = create_app()


@app.route("/")
def home():
    return jsonify([])


@app.after_request
def after_request_func(response):
    response.direct_passthrough = False
    return response


@user_registered.connect_via(app)
def user_registered_sighandler(app, user, confirmation_token, **kwargs):
    default_role = app.security.datastore.find_or_create_role(
        name="user", permissions={"user-read", "user-write"}
    )
    app.security.datastore.add_role_to_user(user, default_role)


if __name__ == '__main__':
    with app.app_context():
        logger.debug("Loading DB..")
        db.create_all()

        # Download the icon sets if they aren't already in the DB
        icons = db.session.query(Icon).count()
        if icons == 0:
            logger.info("Downloading icons...")
            try:
                r = requests.get("https://github.com/brian7704/OpenTAKServer-Installer/raw/master/iconsets.sqlite", stream=True)
                with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "icons.sqlite"), "wb") as f:
                    f.write(r.content)

                def dict_factory(cursor, row):
                    d = {}
                    for idx, col in enumerate(cursor.description):
                        d[col[0]] = row[idx]
                    return d

                con = sqlite3.connect(os.path.join(app.config.get("OTS_DATA_FOLDER"), "icons.sqlite"))
                con.row_factory = dict_factory
                cur = con.cursor()
                rows = cur.execute("SELECT * FROM icons")
                for row in rows:
                    db.session.execute(insert(Icon).values(**row))
                db.session.commit()
            except BaseException as e:
                logger.error("Failed to download icons: {}".format(e))

        if app.config.get("DEBUG"):
            logger.debug("Starting in debug mode")
        else:
            logger.info("Starting in production mode")

        app.security.datastore.find_or_create_role(
            name="user", permissions={"user-read", "user-write"}
        )

        app.security.datastore.find_or_create_role(
            name="administrator", permissions={"administrator"}
        )

        if not app.security.datastore.find_user(username="administrator"):
            logger.info("Creating administrator account. The password is 'password'")
            app.security.datastore.create_user(username="administrator",
                                               password=hash_password("password"), roles=["administrator"])
        db.session.commit()

    tcp_thread = SocketServer(logger, app.app_context(), app.config.get("OTS_TCP_STREAMING_PORT"))
    tcp_thread.start()
    app.tcp_thread = tcp_thread

    ssl_thread = SocketServer(logger, app.app_context(), app.config.get("OTS_SSL_STREAMING_PORT"), True)
    ssl_thread.start()
    app.ssl_thread = ssl_thread

    if app.config.get("OTS_ENABLE_MUMBLE_AUTHENTICATION"):
        try:
            logger.info("Starting Mumble authentication handler")
            mumble_daemon = MumbleIceDaemon(app, logger)
            mumble_daemon.daemon = True
            mumble_daemon.start()
        except BaseException as e:
            logger.error("Failed to enable Mumble authentication: {}".format(e))
            logger.error(traceback.format_exc())
    else:
        logger.info("Mumble authentication handler disabled")

    app.start_time = datetime.now()

    socketio.run(app, host=app.config.get("OTS_LISTENER_ADDRESS"), port=app.config.get("OTS_LISTENER_PORT"),
                 debug=app.config.get("DEBUG"), log_output=app.config.get("DEBUG"))



import os
import traceback

from flask_security import roles_required
from sqlalchemy import or_, update
import pickle

import bleach
from flask import current_app as app, request, Blueprint, jsonify
from flask_security import uia_username_mapper, uia_email_mapper

from opentakserver.extensions import db, logger
from opentakserver.models.Config import ConfigSettings

config_blueprint = Blueprint('config_blueprint', __name__)


def search(query, model, field):
    arg = request.args.get(field)
    if arg:
        arg = bleach.clean(arg)
        return query.where(getattr(model, field) == arg)
    return query


@config_blueprint.route('/api/config')
@roles_required('administrator')
def get_config():
    query = db.session.query(ConfigSettings).filter(
        or_(*[ConfigSettings.key.like("OTS%"), ConfigSettings.key.like("MAIL%")]))
    query = search(query, ConfigSettings, 'key')

    results = {}
    types = []

    settings = db.session.execute(query).all()
    for setting in settings:
        setting = setting[0]

        if type(pickle.loads(setting.value)) not in types:
            types.append(type(pickle.loads(setting.value)))

        if type(pickle.loads(setting.value)) == 'function':
            logger.warning(setting)
        elif setting.type == 'timedelta':
            results[setting.key] = str(pickle.loads(setting.value))
        else:
            results[setting.key] = pickle.loads(setting.value)

    return jsonify({'success': True, 'results': results})


@config_blueprint.route('/api/config', methods=['PATCH'])
@roles_required('administrator')
def change_setting():
    if not request.json:
        return jsonify({'success': False, 'error': 'Empty request'}), 400

    try:
        for setting in request.json:
            setting = bleach.clean(setting)
            config = db.session.execute(db.session.query(ConfigSettings).where(ConfigSettings.key == setting)).first()[0]
            value = request.json.get(setting)

            if value and type(value) is 'str':
                value = bleach.clean(value)

            #  Enable/disable account registration, confirmation, and recovery  based on whether email is enabled
            if setting == 'OTS_ENABLE_EMAIL':
                app.config.update({"SECURITY_REGISTERABLE": value})
                db.session.execute(update(ConfigSettings).filter(ConfigSettings.key == "SECURITY_REGISTERABLE")
                                   .values(value=pickle.dumps(value)))
                app.config.update({"SECURITY_CONFIRMABLE": value})
                db.session.execute(update(ConfigSettings).filter(ConfigSettings.key == "SECURITY_CONFIRMABLE")
                                   .values(value=pickle.dumps(value)))
                app.config.update({"SECURITY_RECOVERABLE": value})
                db.session.execute(update(ConfigSettings).filter(ConfigSettings.key == "SECURITY_RECOVERABLE")
                                   .values(value=pickle.dumps(value)))

                #  Users can always enable authenticator based 2FA. If email is enabled they can also use email based 2FA
                two_factor_methods = ["authenticator"]
                if value:
                    two_factor_methods.append("email")
                app.config.update({"SECURITY_TWO_FACTOR_ENABLED_METHODS": two_factor_methods})
                db.session.execute(
                    update(ConfigSettings).filter(ConfigSettings.key == "SECURITY_TWO_FACTOR_ENABLED_METHODS")
                    .values(value=pickle.dumps(two_factor_methods)))

                #  Users can always log in with usernames. If email is enabled they can also log in with email addresses
                identity_attributes = [{"username": {"mapper": uia_username_mapper, "case_insensitive": False}}]
                if value:
                    identity_attributes.append({"email": {"mapper": uia_email_mapper, "case_insensitive": True}})
                app.config.update({"SECURITY_USER_IDENTITY_ATTRIBUTES": identity_attributes})
                db.session.execute(
                    update(ConfigSettings).filter(ConfigSettings.key == "SECURITY_USER_IDENTITY_ATTRIBUTES")
                    .values(value=pickle.dumps(identity_attributes)))

                config.value = pickle.dumps(value)
                app.config.update({setting: value})
                db.session.add(config)

            elif setting == "OTS_ENABLE_MUMBLE_AUTHENTICATION" or setting.startswith("OTS_AIRPLANES_LIVE") or \
                    setting.startswith("MAIL"):

                config.value = pickle.dumps(value)
                app.config.update({setting: value})
                db.session.add(config)

        db.session.commit()
    except BaseException as e:
        db.session.rollback()
        logger.error(request.json)
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 500

    return jsonify({'success': True})



import json
import re
import traceback
from threading import Thread

import bleach
import sqlalchemy.exc
from sqlalchemy import exc, insert, update
from bs4 import BeautifulSoup
import pika

from opentakserver.extensions import socketio
from opentakserver.functions import datetime_from_iso8601_string
from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.Alert import Alert
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.CoT import CoT
from opentakserver.models.EUD import EUD
from opentakserver.models.GeoChat import GeoChat
from opentakserver.models.Icon import Icon
from opentakserver.models.RBLine import RBLine
from opentakserver.models.Team import Team
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.ZMIST import ZMIST
from opentakserver.models.Point import Point
from opentakserver.models.Marker import Marker


class CoTController:
    def __init__(self, context, logger, db, socketio):
        self.context = context
        self.logger = logger
        self.db = db
        self.socketio = socketio

        self.online_euds = {}
        self.online_callsigns = {}
        self.exchanges = []

        # RabbitMQ
        try:
            self.rabbit_connection = pika.SelectConnection(pika.ConnectionParameters(self.context.app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")),
                                                           self.on_connection_open)
            self.rabbit_channel = None
            self.iothread = Thread(target=self.rabbit_connection.ioloop.start)
            self.iothread.daemon = True
            self.iothread.start()
            self.is_consuming = False
        except BaseException as e:
            self.logger.error("cot_controller - Failed to connect to rabbitmq: {}".format(e))
            return

    def on_connection_open(self, connection):
        self.rabbit_connection.channel(on_open_callback=self.on_channel_open)
        self.rabbit_connection.add_on_close_callback(self.on_close)

    def on_channel_open(self, channel):
        self.rabbit_channel = channel
        self.rabbit_channel.queue_declare(queue='cot_controller')
        self.rabbit_channel.exchange_declare(exchange='cot_controller', exchange_type='fanout')
        self.rabbit_channel.queue_bind(exchange='cot_controller', queue='cot_controller')
        self.rabbit_channel.basic_consume(queue='cot_controller', on_message_callback=self.on_message, auto_ack=True)
        self.rabbit_channel.add_on_close_callback(self.on_close)

    def on_close(self, channel, error):
        self.logger.error("cot_controller closing RabbitMQ connection: {}".format(error))

    def parse_device_info(self, uid, soup, event):
        link = event.find('link')
        fileshare = event.find('fileshare')

        # Don't parse server generated messages
        with self.context:
            if uid == self.context.app.config.get("OTS_NODE_ID"):
                return

        callsign = None
        phone_number = None

        if uid not in self.online_euds and not uid.endswith('ping'):
            takv = event.find('takv')
            if takv:
                device = takv.attrs['device'] if 'device' in takv.attrs else ""
                os = takv.attrs['os'] if 'os' in takv.attrs else ""
                platform = takv.attrs['platform'] if 'platform' in takv.attrs else ""
                version = takv.attrs['version'] if 'version' in takv.attrs else ""

                contact = event.find('contact')
                if contact:
                    if 'callsign' in contact.attrs:
                        callsign = contact.attrs['callsign']

                        if uid not in self.online_euds:
                            self.online_euds[uid] = {'cot': str(soup), 'callsign': callsign}

                        if callsign not in self.online_callsigns:
                            self.online_callsigns[callsign] = {'uid': uid, 'cot': soup}

                        # Declare a RabbitMQ Queue for this uid and join the 'dms' and 'cot' exchanges
                        if self.rabbit_channel and self.rabbit_channel.is_open and platform != "OpenTAK ICU":
                            self.rabbit_channel.queue_bind(exchange='dms', queue=uid, routing_key=uid)
                            self.rabbit_channel.queue_bind(exchange='chatrooms', queue=uid,
                                                           routing_key='All Chat Rooms')

                            for eud in self.online_euds:
                                self.rabbit_channel.basic_publish(exchange='dms',
                                                                  routing_key=uid,
                                                                  body=json.dumps({'cot': str(self.online_euds[eud]['cot']),
                                                                                   'uid': None}))

                    if 'phone' in contact.attrs:
                        phone_number = contact.attrs['phone']

                with self.context:
                    group = event.find('__group')
                    team = Team()

                    if group:
                        # Declare an exchange for each group and bind the callsign's queue
                        if self.rabbit_channel.is_open and group.attrs['name'] not in self.exchanges:
                            self.logger.debug("Declaring exchange {}".format(group.attrs['name']))
                            self.rabbit_channel.exchange_declare(exchange=group.attrs['name'])
                            self.rabbit_channel.queue_bind(queue=uid, exchange='chatrooms', routing_key=group.attrs['name'])
                            self.exchanges.append(group.attrs['name'])

                        team.name = bleach.clean(group.attrs['name'])

                        try:
                            chatroom = self.db.session.execute(self.db.session.query(Chatroom)
                                                               .filter(Chatroom.name == team.name)).first()[0]
                            team.chatroom_id = chatroom.id
                        except TypeError:
                            chatroom = None

                        try:
                            self.db.session.add(team)
                            self.db.session.commit()
                        except sqlalchemy.exc.IntegrityError:
                            self.db.session.rollback()
                            team = self.db.session.execute(self.db.session.query(Team)
                                                           .filter(Team.name == group.attrs['name'])).first()[0]
                            if not team.chatroom_id and chatroom:
                                team.chatroom_id = chatroom.id
                                self.db.session.execute(update(Team).filter(Team.name).values(**team))

                    try:
                        eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=uid)).first()[0]
                    except:
                        eud = EUD()

                    eud.uid = uid
                    eud.callsign = callsign
                    eud.device = device
                    eud.os = os
                    eud.platform = platform
                    eud.version = version
                    eud.phone_number = phone_number
                    eud.last_event_time = datetime_from_iso8601_string(event.attrs['start'])
                    eud.last_status = 'Connected'

                    if group:
                        eud.team_id = team.id
                        eud.team_role = bleach.clean(group.attrs['role'])

                    self.db.session.add(eud)
                    self.db.session.commit()

                    socketio.emit('eud', eud.to_json(), namespace='/socket.io')

        # Update the CoT stored in memory which contains the new stale time
        elif event.find('takv'):
            self.online_euds[uid]['cot'] = str(soup)

    def insert_cot(self, soup, event, uid):
        try:
            sender_callsign = self.online_euds[uid]['callsign']
        except:
            sender_callsign = 'server'

        start = datetime_from_iso8601_string(event.attrs['start'])
        stale = datetime_from_iso8601_string(event.attrs['stale'])
        timestamp = datetime_from_iso8601_string(event.attrs['time'])

        with self.context:
            res = self.db.session.execute(insert(CoT).values(
                how=event.attrs['how'], type=event.attrs['type'], sender_callsign=sender_callsign,
                sender_uid=uid, timestamp=timestamp, xml=str(soup), start=start, stale=stale
            ))

            self.db.session.commit()
            return res.inserted_primary_key[0]

    def parse_point(self, event, uid, cot_id):
        # hae = Height above the WGS ellipsoid in meters
        # ce = Circular 1-sigma or a circular area about the location in meters
        # le = Linear 1-sigma error or an altitude range about the location in meters
        point = event.find('point')
        if point and not point.attrs['lat'].startswith('999'):
            p = Point()
            p.uid = event.attrs['uid']
            p.device_uid = uid
            p.ce = point.attrs['ce']
            p.hae = point.attrs['hae']
            p.le = point.attrs['le']
            p.latitude = float(point.attrs['lat'])
            p.longitude = float(point.attrs['lon'])
            p.timestamp = datetime_from_iso8601_string(event.attrs['time'])
            p.cot_id = cot_id

            # We only really care about the rest of the data if there's a valid lat/lon
            if p.latitude == 0 and p.longitude == 0:
                return None

            track = event.find('track')
            if track:
                if 'course' in track.attrs and track.attrs['course'] != "9999999.0":
                    p.course = track.attrs['course']
                else:
                    p.course = 0

                if 'speed' in track.attrs and track.attrs['speed'] != "9999999.0":
                    p.speed = track.attrs['speed']
                else:
                    p.speed = 0

            # For TAK ICU and OpenTAK ICU CoT's with bearing from the compass
            sensor = event.find('sensor')
            if sensor:
                if 'azimuth' in sensor.attrs:
                    p.azimuth = sensor.attrs['azimuth']
                # Camera's field of view
                if 'fov' in sensor.attrs:
                    p.fov = sensor.attrs['fov']

            precision_location = event.find('precisionlocation')
            if precision_location and 'geolocationsrc' in precision_location.attrs:
                p.location_source = precision_location.attrs['geolocationsrc']
            elif precision_location and 'altsrc' in precision_location.attrs:
                p.location_source = precision_location.attrs['altsrc']
            elif event.attrs['how'] == 'm-g':
                p.location_source = 'GPS'

            status = event.find('status')
            if status:
                if 'battery' in status.attrs:
                    p.battery = status.attrs['battery']

            with self.context:
                res = self.db.session.execute(insert(Point).values(
                    uid=p.uid, device_uid=p.device_uid, ce=p.ce, hae=p.hae, le=p.le, latitude=p.latitude,
                    longitude=p.longitude, timestamp=p.timestamp, cot_id=cot_id, location_source=p.location_source,
                    course=p.course, speed=p.speed, battery=p.battery, fov=p.fov, azimuth=p.azimuth)
                )

                self.db.session.commit()
                # Get the point from the DB with its related CoT
                p = self.db.session.execute(
                    self.db.session.query(Point).filter(Point.id == res.inserted_primary_key[0])).first()[0]

                # This CoT is a position update for an EUD, send it to socketio clients, so it can be seen on the UI map
                # OpenTAK ICU position updates don't include the <takv> tag, but we still want to send the updated position
                # to the UI's map
                if event.find('takv') or event.find("__video"):
                    socketio.emit('point', p.to_json(), namespace='/socket.io')

                return res.inserted_primary_key[0]

    def parse_geochat(self, event, cot_id, point_pk):
        chat = event.find('__chat')
        if chat:
            chat_group = event.find('chatgrp')
            remarks = event.find('remarks')

            chatroom = Chatroom()

            chatroom.name = chat.attrs['chatroom']
            chatroom.id = chat.attrs['id']
            chatroom.parent = chat.attrs['parent']

            with self.context:
                try:
                    self.db.session.add(chatroom)
                    self.db.session.commit()
                except exc.IntegrityError:
                    self.db.session.rollback()

            geochat = GeoChat()

            geochat.uid = event.attrs['uid']
            geochat.chatroom_id = chat.attrs['id']
            geochat.sender_uid = chat_group.attrs['uid0']
            geochat.remarks = remarks.text
            geochat.timestamp = datetime_from_iso8601_string(remarks.attrs['time'])
            geochat.point_id = point_pk
            geochat.cot_id = cot_id

            with self.context:
                try:

                    self.db.session.add(geochat)
                    self.db.session.commit()
                except exc.IntegrityError:
                    # TODO: Check if remarks can be edited and if so do an update here
                    self.db.session.rollback()

            for attr in chat_group.attrs:
                if attr.startswith("uid") and attr != 'uid':

                    if chat.attrs['groupOwner'].lower() == 'true' and attr == 'uid0':
                        with self.context:
                            self.db.session.execute(update(Chatroom).where(Chatroom.id == chat.attrs['id'])
                                                    .values(group_owner=chat_group.attrs[attr]))
                            self.db.session.commit()

                    chatroom_uid = ChatroomsUids()
                    chatroom_uid.chatroom_id = chat.attrs['id']
                    chatroom_uid.uid = chat_group.attrs[attr]

                    with self.context:
                        try:
                            self.db.session.add(chatroom_uid)
                            self.db.session.commit()
                            self.logger.debug(
                                "add {} to chatroom {}".format(chatroom_uid.uid, chatroom_uid.chatroom_id))
                        except exc.IntegrityError:
                            self.db.session.rollback()

    def parse_video(self, event, cot_pk):
        video = event.find("__video")
        if video:
            self.logger.debug("Got video stream")
            connection_entry = video.find('ConnectionEntry')
            if not connection_entry:
                return

            path = connection_entry.attrs['path']
            if path.startswith("/"):
                path = path[1:]

            v = VideoStream()
            v.network_timeout = connection_entry.attrs['networkTimeout']
            v.uid = connection_entry.attrs['uid']
            v.path = path
            v.protocol = connection_entry.attrs['protocol']
            v.buffer_time = connection_entry.attrs['bufferTime']
            v.port = connection_entry.attrs['port']
            v.rover_port = connection_entry.attrs['roverPort']
            v.rtsp_reliable = connection_entry.attrs['rtspReliable']
            v.ignore_embedded_klv = (connection_entry.attrs['ignoreEmbeddedKLV'].lower() == 'true')
            v.alias = connection_entry.attrs['alias']
            v.cot_id = cot_pk
            v.generate_xml(connection_entry.attrs['address'])

            with self.context:
                try:
                    self.db.session.add(v)
                    self.db.session.commit()
                    self.logger.debug("Added video")
                except exc.IntegrityError as e:
                    self.db.session.rollback()
                    self.db.session.execute(update(VideoStream).where(VideoStream.uid == connection_entry.attrs['uid'])
                                            .values(network_timeout=connection_entry.attrs['networkTimeout'],
                                                    protocol=connection_entry.attrs['protocol'],
                                                    buffer_time=connection_entry.attrs['bufferTime'],
                                                    address=connection_entry.attrs['address'],
                                                    port=connection_entry.attrs['port'],
                                                    rover_port=connection_entry.attrs['roverPort'],
                                                    rtsp_reliable=connection_entry.attrs['rtspReliable'],
                                                    ignore_embedded_klv=(connection_entry.attrs[
                                                                             'ignoreEmbeddedKLV'].lower() == 'true'),
                                                    alias=connection_entry.attrs['alias'],
                                                    xml=v.xml))

                    self.db.session.commit()

    def parse_alert(self, event, uid, point_pk, cot_pk):
        emergency = event.find('emergency')
        if emergency:
            if 'type' in emergency.attrs:
                emergency_type = emergency.attrs['type']
                alert = Alert()
                alert.sender_uid = uid
                alert.uid = event.attrs['uid']
                alert.start_time = datetime_from_iso8601_string(event.attrs['start'])
                alert.point_id = point_pk
                alert.alert_type = emergency_type
                alert.cot_id = cot_pk

                with self.context:
                    self.db.session.add(alert)
                    self.db.session.commit()
                    socketio.emit('alert', alert.to_json(), namespace='/socket.io')
            elif 'cancel' in emergency.attrs:
                with self.context:
                    try:
                        alert = self.db.session.execute(
                            Alert.query.filter(Alert.cancel_time == None, Alert.sender_uid == uid).order_by(
                                Alert.start_time.desc())).first()[0]
                        alert.cancel_time = datetime_from_iso8601_string(event.attrs['start'])
                        self.db.session.commit()
                        socketio.emit('alert', alert.to_json(), namespace='/socket.io')
                    except BaseException as e:
                        self.logger.error("Failed to set alert cancel time: {}".format(e))

    def parse_casevac(self, event, uid, point_pk, cot_pk):
        medevac = event.find('_medevac_')
        if medevac:
            zmist = medevac.find('zMist')
            with self.context:
                for a in medevac.attrs:
                    if medevac.attrs[a].lower() == 'true':
                        medevac.attrs[a] = True
                    elif medevac.attrs[a].lower() == 'false':
                        medevac.attrs[a] = False

                try:
                    res = self.db.session.execute(
                        insert(CasEvac).values(timestamp=datetime_from_iso8601_string(event.attrs['start']),
                                               sender_uid=uid, uid=event.attrs['uid'],
                                               point_id=point_pk, cot_id=cot_pk,
                                               **medevac.attrs))
                    casevac_pk = res.inserted_primary_key[0]

                    if zmist:
                        self.db.session.execute(insert(ZMIST).values(casevac_uid=event.attrs['uid'], **zmist.attrs))
                except exc.IntegrityError as e:
                    self.db.session.rollback()
                    self.db.session.execute(update(CasEvac).where(CasEvac.uid == event.attrs['uid'])
                                            .values(**medevac.attrs))

                    if zmist:
                        self.db.session.execute(
                            update(ZMIST).where(CasEvac.uid == event.attrs['uid']).values(**zmist.attrs))
                self.db.session.commit()

    def parse_marker(self, event, uid, point_pk, cot_pk):
        if ((re.match("^a-[f|h|u|p|a|n|s|j|k]-[Z|P|A|G|S|U|F]", event.attrs['type']) or
             # Spot map
             re.match("^b-m-p", event.attrs['type'])) and
                # Don't worry about EUD location updates
                not event.find('takv') and
                # Ignore video streams from sources like OpenTAK ICU
                event.attrs['type'] != 'b-m-p-s-p-loc'):

            try:
                marker = Marker()
                marker.uid = event.attrs['uid']
                marker.affiliation = self.get_affiliation(event.attrs['type'])
                marker.battle_dimension = self.get_battle_dimension(event.attrs['type'])

                marker.mil_std_2525c = "s"
                cot_type_list = event.attrs['type'].split("-")
                cot_type_list.pop(0)  # this should always be letter a
                affiliation = cot_type_list.pop(0)
                battle_dimension = cot_type_list.pop(0)
                marker.mil_std_2525c += affiliation
                marker.mil_std_2525c += battle_dimension
                marker.mil_std_2525c += "-"

                for letter in cot_type_list:
                    if letter.isupper():
                        marker.mil_std_2525c += letter.lower()

                while len(marker.mil_std_2525c) < 10:
                    marker.mil_std_2525c += "-"

                detail = event.find('detail')
                icon = None

                if detail:
                    for tag in detail:
                        if 'readiness' in tag.attrs:
                            marker.readiness = tag.attrs['readiness'] == "true"
                        if 'argb' in tag.attrs:
                            marker.argb = tag.attrs['argb']
                            marker.color_hex = marker.color_to_hex()
                        if 'callsign' in tag.attrs:
                            marker.callsign = tag.attrs['callsign']
                        if 'iconsetpath' in tag.attrs:
                            marker.iconset_path = tag.attrs['iconsetpath']
                            if marker.iconset_path.lower().endswith('.png'):
                                with self.context:
                                    filename = marker.iconset_path.split("/")[-1]

                                    try:
                                        icon = self.db.session.execute(self.db.session.query(Icon)
                                                                       .filter(Icon.filename == filename)).first()[0]
                                        marker.icon_id = icon.id
                                    except:
                                        icon = self.db.session.execute(self.db.session.query(Icon)
                                        .filter(
                                            Icon.filename == 'marker-icon.png')).first()[0]
                                        marker.icon_id = icon.id
                            elif not marker.mil_std_2525c:
                                with self.context:
                                    icon = self.db.session.execute(self.db.session.query(Icon)
                                                                   .filter(Icon.filename == 'marker-icon.png')).first()[
                                        0]
                                    marker.icon_id = icon.id

                        if 'altsrc' in tag.attrs:
                            marker.location_source = tag.attrs['altsrc']

                link = event.find('link')
                if link:
                    marker.parent_callsign = link.attrs['parent_callsign'] if 'parent_callsign' in link.attrs else None
                    marker.production_time = link.attrs['production_time'] if 'production_time' in link.attrs else None
                    marker.relation = link.attrs['relation'] if 'relation' in link.attrs else None
                    marker.relation_type = link.attrs['relation_type'] if 'relation_type' in link.attrs else None
                    marker.parent_uid = link.attrs['uid'] if 'uid' in link.attrs else None

                marker.point_id = point_pk
                marker.cot_id = cot_pk

                with self.context:
                    try:
                        self.db.session.add(marker)
                        self.db.session.commit()
                        self.logger.debug('added marker')
                    except exc.IntegrityError:
                        self.db.session.rollback()
                        self.db.session.execute(
                            update(Marker).where(Marker.uid == marker.uid).values(point_id=marker.point_id,
                                                                                  icon_id=marker.icon_id,
                                                                                  **marker.serialize()))
                        self.db.session.commit()
                        self.logger.debug('updated marker')
                        marker = self.db.session.execute(self.db.session.query(Marker)
                                                         .filter(Marker.uid == marker.uid)).first()[0]

                    socketio.emit('marker', marker.to_json(), namespace='/socket.io')

            except BaseException as e:
                self.logger.error("Failed to parse marker: {}".format(e))
                self.logger.error(traceback.format_exc())

    def parse_rbline(self, event, uid, point_pk, cot_pk):
        if re.match("^u-rb", event.attrs['type']):
            self.logger.debug("Got an R&B line")
            rb_line = RBLine()

            detail = event.find('detail')
            if detail:
                rb_line.uid = event.attrs['uid']
                rb_line.sender_uid = uid
                rb_line.timestamp = datetime_from_iso8601_string(event.attrs['start'])
                rb_line.point_id = point_pk
                rb_line.cot_id = cot_pk

                for tag in detail:
                    if tag.name == 'range':
                        rb_line.range = tag.attrs['value']
                    if tag.name == 'bearing':
                        rb_line.bearing = tag.attrs['value']
                    if tag.name == 'inclination':
                        rb_line.inclination = tag.attrs['value']
                    if tag.name == 'anchorUID':
                        rb_line.anchor_uid = tag.attrs['value']
                    if tag.name == 'rangeUnits':
                        rb_line.range_units = tag.attrs['value']
                    if tag.name == 'bearingUnits':
                        rb_line.bearing_units = tag.attrs['value']
                    if tag.name == 'northRef':
                        rb_line.north_ref = tag.attrs['value']
                    if tag.name == 'color':
                        rb_line.color = tag.attrs['value']
                        rb_line.color_hex = rb_line.color_to_hex()
                    if tag.name == 'contact':
                        rb_line.callsign = tag.attrs['callsign']
                    if tag.name == 'strokeColor':
                        rb_line.stroke_color = tag.attrs['value']
                    if tag.name == 'strokeWeight':
                        rb_line.stroke_weight = tag.attrs['value']
                    if tag.name == 'labels_on':
                        rb_line.labels_on = (tag.attrs['value'] == 'true')

                with self.context:

                    start_point = \
                        self.db.session.execute(self.db.session.query(Point).filter(Point.id == point_pk)).first()[0]
                    end_point = rb_line.calc_end_point(start_point)
                    rb_line.end_latitude = end_point['latitude']
                    rb_line.end_longitude = end_point['longitude']

                    try:
                        self.db.session.add(rb_line)
                        self.db.session.commit()
                        self.logger.debug("Inserted new R&B line: {}".format(rb_line.uid))
                    except exc.IntegrityError:
                        self.db.session.rollback()
                        self.db.session.execute(update(RBLine).where(RBLine.uid == rb_line.uid)
                                                .values(**rb_line.serialize()))
                        self.db.session.commit()
                        self.logger.debug('Updated R&B line: {}'.format(rb_line.uid))

                    rb_line.point = start_point
                    socketio.emit("rb_line", rb_line.to_json(), namespace='/socket.io')

    def rabbitmq_routing(self, event, data):
        # RabbitMQ Routing
        chat = event.find("__chat")
        destinations = event.find_all('dest')

        if chat and 'chatroom' in chat.attrs and chat.attrs['chatroom'] == 'All Chat Rooms':
            self.rabbit_channel.basic_publish(exchange='chatrooms', routing_key='All Chat Rooms', body=json.dumps(data))

        elif destinations:
            for destination in destinations:
                # ATAK and WinTAK use callsign, iTAK uses uid
                if 'callsign' in destination.attrs:
                    uid = self.online_callsigns[destination.attrs['callsign']]['uid']
                elif 'uid' in destination.attrs:
                    uid = destination.attrs['uid']
                else:
                    continue

                self.rabbit_channel.basic_publish(exchange='dms',
                                                  routing_key=uid,
                                                  body=json.dumps(data))

        # If no destination or callsign is specified, broadcast to all TAK clients
        elif self.rabbit_channel and self.rabbit_channel.is_open:
            self.rabbit_channel.basic_publish(exchange='cot', routing_key="", body=json.dumps(data))

        # Do nothing because the RabbitMQ channel hasn't opened yet or has closed
        else:
            self.logger.debug("Not publishing, channel closed")

    def get_affiliation(self, type):
        if re.match("^t-", type):
            return self.get_tasking(type)
        if re.match("^a-f-", type):
            return "friendly"
        if re.match("^a-h-", type):
            return "hostile"
        if re.match("^a-u-", type):
            return "unknown"
        if re.match("^a-p-", type):
            return "pending"
        if re.match("^a-a-", type):
            return "assumed"
        if re.match("^a-n-", type):
            return "neutral"
        if re.match("^a-s-", type):
            return "suspect"
        if re.match("^a-j-", type):
            return "joker"
        if re.match("^a-k-", type):
            return "faker"

    def get_tasking(self, type):
        if re.match("^t-x-f", type):
            return "remarks"
        if re.match("^t-x-s", type):
            return "state/sync"
        if re.match("^t-s", type):
            return "required"
        if re.match("^t-z", type):
            return "cancel"
        if re.match("^t-x-c-c", type):
            return "commcheck"
        if re.match("^t-x-c-g-d", type):
            return "dgps"
        if re.match("^t-k-d", type):
            return "destroy"
        if re.match("^t-k-i", type):
            return "investigate"
        if re.match("^t-k-t", type):
            return "target"
        if re.match("^t-k", type):
            return "strike"
        if re.match("^t-", type):
            return "tasking"

    def get_battle_dimension(self, type):
        if re.match("^a-.-A", type):
            return "airborne"
        if re.match("^a-.-G", type):
            return "ground"
        if re.match("^a-.-G-I", type):
            return "installation"
        if re.match("^a-.-S", type):
            return "surface/sea"
        if re.match("^a-.-U", type):
            return "subsurface"

    def parse_type(self, type):
        if re.match("^a-.-G-I", type):
            return "installation"
        if re.match("^a-.-G-E-V", type):
            return "vehicle"
        if re.match("^a-.-G-E", type):
            return "equipment"
        if re.match("^a-.-A-W-M-S", type):
            return "sam"
        if re.match("^a-.-A-M-F-Q-r", type):
            return "uav"

    def on_message(self, unused_channel, basic_deliver, properties, body):
        try:
            body = json.loads(body)
            soup = BeautifulSoup(body['cot'], 'xml')
            event = soup.find('event')
            self.logger.info(event)
            if event:
                self.parse_device_info(body['uid'], soup, event)
                cot_pk = self.insert_cot(soup, event, body['uid'])
                point_pk = self.parse_point(event, body['uid'], cot_pk)
                self.parse_geochat(event, cot_pk, point_pk)
                self.parse_video(event, cot_pk)
                self.parse_alert(event, body['uid'], point_pk, cot_pk)
                self.parse_casevac(event, body['uid'], point_pk, cot_pk)
                self.parse_marker(event, body['uid'], point_pk, cot_pk)
                self.parse_rbline(event, body['uid'], point_pk, cot_pk)
                self.rabbitmq_routing(event, body)

                # EUD went offline
                if event.attrs['type'] == 't-x-d-d':
                    link = event.find('link')

                    try:
                        with self.context:
                            eud = self.db.session.execute(self.db.select(EUD).filter_by(uid=body['uid'])).scalar_one()
                            eud.last_event_time = datetime_from_iso8601_string(event.attrs['start'])
                            eud.last_status = 'Disconnected'
                            self.db.session.commit()
                            self.logger.debug("Updated {}".format(body['uid']))
                            eud_json = eud.to_json()
                            # The first time an EUD connects but doesn't have a location.
                            # Tells the UI what kind of EUD this is, ie ATAK/WinTAK/iTAK or OpenTAK ICU
                            if not eud_json['last_point']:
                                eud_json['type'] = event.attrs['type']
                            socketio.emit('eud', eud.to_json(), namespace='/socket.io')
                    except BaseException as e:
                        self.logger.error("Failed to update EUD: {}".format(e))

                    self.online_euds.pop(link.attrs['uid'])
        except BaseException as e:
            self.logger.error(traceback.format_exc())



from datetime import datetime

ISO8601_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
ISO8601_FORMAT_NO_MICROSECONDS = "%Y-%m-%dT%H:%M:%SZ"


def datetime_from_iso8601_string(datetime_string):
    try:
        return datetime.strptime(datetime_string, ISO8601_FORMAT)
    except ValueError:
        return datetime.strptime(datetime_string, ISO8601_FORMAT_NO_MICROSECONDS)


def iso8601_string_from_datetime(datetime_object):
    return datetime_object.strftime("%Y-%m-%dT%H:%M:%S.%f"[:-3] + "Z")



from opentakserver.extensions import db
from sqlalchemy import String, BLOB
from sqlalchemy.orm import Mapped, mapped_column


class ConfigSettings(db.Model):
    __tablename__ = "config"

    key: Mapped[str] = mapped_column(String, primary_key=True)
    type: Mapped[str] = mapped_column(String)
    value: Mapped[bytes] = mapped_column(BLOB)



from dataclasses import dataclass
from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class EUD(db.Model):
    __tablename__ = "euds"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String, nullable=False, unique=True)
    callsign: Mapped[str] = mapped_column(String, nullable=True)
    device: Mapped[str] = mapped_column(String, nullable=True)
    os: Mapped[str] = mapped_column(String, nullable=True)
    platform: Mapped[str] = mapped_column(String, nullable=True)
    version: Mapped[str] = mapped_column(String, nullable=True)
    phone_number: Mapped[int] = mapped_column(Integer, nullable=True)
    last_event_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    last_status: Mapped[str] = mapped_column(String, nullable=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey("user.id"), nullable=True)
    team_id: Mapped[int] = mapped_column(Integer, ForeignKey("teams.id"), nullable=True)
    team_role: Mapped[str] = mapped_column(String, nullable=True)
    points = relationship("Point", back_populates="eud")
    cots = relationship("CoT", back_populates="eud")
    casevacs = relationship("CasEvac", back_populates="eud")
    geochats = relationship("GeoChat", back_populates="eud")
    chatroom_uid = relationship("ChatroomsUids", back_populates="eud")
    user: Mapped["User"] = relationship(back_populates="euds")
    alert = relationship("Alert", back_populates="eud")
    data_packages = relationship("DataPackage", back_populates="eud", uselist=False)
    certificate = relationship("Certificate", back_populates="eud", uselist=False)
    markers = relationship("Marker", back_populates="eud")
    rb_lines = relationship("RBLine", back_populates="eud")
    team = relationship("Team", back_populates="euds")

    def serialize(self):
        return {
            'uid': self.uid,
            'callsign': self.callsign,
            'device': self.device,
            'os': self.os,
            'platform': self.platform,
            'version': self.version,
            'phone_number': self.phone_number,
            'last_event_time': self.last_event_time,
            'last_status': self.last_status,
            'user_id': self.user_id,
            'team_id': self.team_id,
            'team_role': self.team_role
        }

    def to_json(self, include_data_packages=True):
        config_datapackage_hash = None
        if self.certificate and self.certificate.data_package:
            config_datapackage_hash = self.certificate.data_package.hash
        return {
            'uid': self.uid,
            'callsign': self.callsign,
            'device': self.device,
            'os': self.os,
            'platform': self.platform,
            'version': self.version,
            'phone_number': self.phone_number,
            'last_event_time': iso8601_string_from_datetime(self.last_event_time) if self.last_event_time else None,
            'last_status': self.last_status,
            'username': self.user.username if self.user else None,
            'last_point': self.points[-1].to_json() if self.points else None,
            'team': self.team.name if self.team else None,
            'team_color': self.team.get_team_color() if self.team else None,
            'team_role': self.team_role,
            'data_packages': self.data_packages.to_json(False) if include_data_packages and self.data_packages else None,
            'config_hash': config_datapackage_hash
        }



import base64
from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import Integer, String, LargeBinary
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class Icon(db.Model):
    __tablename__ = "icons"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    iconset_uid: Mapped[str] = mapped_column(String, nullable=True)
    filename: Mapped[str] = mapped_column(String, nullable=True)
    groupName: Mapped[str] = mapped_column(String, nullable=True)
    type2525b: Mapped[str] = mapped_column(String, nullable=True)
    useCnt: Mapped[int] = mapped_column(Integer, nullable=True)
    bitmap: Mapped[bytes] = mapped_column(LargeBinary, nullable=True)
    shadow: Mapped[bytes] = mapped_column(LargeBinary, nullable=True)
    markers = relationship("Marker", back_populates="icon")

    def serialize(self):
        return {
            'iconset_uid': self.iconset_uid,
            'filename': self.filename,
            'groupName': self.groupName,
            'type2525b': self.type2525b,
            'useCnt': self.useCnt,
            'bitmap': 'data:image/png;base64,{}'.format(base64.b64encode(self.bitmap).decode('utf-8')) if self.bitmap else None,
            'shadow': 'data:image/png;base64,{}'.format(base64.b64encode(self.shadow).decode('utf-8')) if self.shadow else None
        }

    def to_json(self):
        return self.serialize()



import math
from dataclasses import dataclass
from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Float, Boolean, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from pygc import great_circle

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class RBLine(db.Model):
    __tablename__ = "rb_lines"

    # type = u-rb-a
    # how = h-e

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    sender_uid: Mapped[str] = mapped_column(String, ForeignKey("euds.uid"))
    uid: Mapped[str] = mapped_column(String, unique=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)

    range: Mapped[float] = mapped_column(Float)
    bearing: Mapped[float] = mapped_column(Float)
    inclination: Mapped[float] = mapped_column(Float, nullable=True)
    anchor_uid: Mapped[str] = mapped_column(String, nullable=True)
    range_units: Mapped[int] = mapped_column(Integer, nullable=True)
    bearing_units: Mapped[int] = mapped_column(Integer, nullable=True)
    north_ref: Mapped[int] = mapped_column(Integer, nullable=True)
    color: Mapped[int] = mapped_column(Integer, nullable=True)
    color_hex: Mapped[str] = mapped_column(String, nullable=True)
    callsign: Mapped[str] = mapped_column(String, nullable=True)
    stroke_color: Mapped[int] = mapped_column(Integer, nullable=True)
    stroke_weight: Mapped[float] = mapped_column(Float, nullable=True)
    stroke_style: Mapped[str] = mapped_column(String, nullable=True)
    labels_on: Mapped[bool] = mapped_column(Boolean, nullable=True)
    point_id: Mapped[int] = mapped_column(Integer, ForeignKey("points.id"), nullable=True)
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id"), nullable=True)
    end_latitude: Mapped[float] = mapped_column(Float, nullable=True)
    end_longitude: Mapped[float] = mapped_column(Float, nullable=True)
    point = relationship("Point", back_populates="rb_line")
    cot = relationship("CoT", back_populates="rb_line")
    eud = relationship("EUD", back_populates="rb_lines")

    range_unit_names = ['standard', 'metric', 'nautical']
    bearing_unit_names = ['degrees', 'mils']
    north_ref_names = ['true', 'magnetic', 'grid']

    def color_to_hex(self):
        if self.color:
            return format(int(self.color) & 0xFFFFFFFF, '08X')

    def calc_end_point(self, start_point):
        if not int(self.bearing_units):
            azimuth = float(self.bearing)
        else:
            azimuth = math.degrees(float(self.bearing))

        return great_circle(distance=float(self.range), azimuth=azimuth, latitude=start_point.latitude,
                            longitude=start_point.longitude)

    def serialize(self):
        return {
            'uid': self.uid,
            'timestamp': self.timestamp,
            'range': self.range,
            'inclination': self.inclination,
            'anchor_uid': self.anchor_uid,
            'range_units': self.range_units,
            'bearing_units': self.bearing_units,
            'north_ref': self.north_ref,
            'color': self.color,
            'color_hex': self.color_hex,
            'stroke_weight': self.stroke_weight,
            'stroke_style': self.stroke_style,
            'labels_on': self.labels_on,
            'end_latitude': self.end_latitude,
            'end_longitude': self.end_longitude
        }

    def to_json(self):
        return {
            'uid': self.uid,
            'timestamp': iso8601_string_from_datetime(self.timestamp),
            'range': self.range,
            'inclination': self.inclination,
            'anchor_uid': self.anchor_uid,
            'range_units': self.range_units,
            'range_unit_name': self.range_unit_names[int(self.range_units)] if self.range_units else None,
            'bearing_units': self.bearing_units,
            'bearing_unit_name': self.bearing_unit_names[int(self.bearing_units)] if self.bearing else None,
            'north_ref': self.north_ref,
            'north_ref_name': self.north_ref_names[int(self.north_ref)] if self.north_ref else None,
            'color': self.color,
            'color_hex': self.color_hex,
            'stroke_weight': self.stroke_weight,
            'stroke_style': self.stroke_style,
            'labels_on': self.labels_on,
            'point': self.point.to_json() if self.point else None,
            'end_latitude': self.end_latitude,
            'end_longitude': self.end_longitude
        }



import os
import socket
import ssl
from threading import Thread

from opentakserver.controllers.client_controller import ClientController


class SocketServer(Thread):
    def __init__(self, logger, app_context=None, port=8088, ssl_server=False):
        super().__init__()

        self.logger = logger
        self.port = port
        self.ssl = ssl_server
        self.shutdown = False
        self.daemon = True
        self.socket = None
        self.clients = []
        self.app_context = app_context

    def run(self):
        if self.ssl:
            self.socket = self.launch_ssl_server()
        elif self.app_context.app.config.get("OTS_ENABLE_TCP_STREAMING_PORT"):
            self.socket = self.launch_tcp_server()
        else:
            self.logger.info("TCP connections are disabled")
            return

        self.socket.settimeout(1.0)

        while not self.shutdown:
            try:
                sock, addr = self.socket.accept()
                if self.ssl:
                    self.logger.info("New SSL connection from {}".format(addr[0]))
                else:
                    self.logger.info("New TCP connection from {}".format(addr[0]))

                new_thread = ClientController(addr[0], addr[1], sock, self.logger, self.app_context.app, self.ssl)
                new_thread.daemon = True
                new_thread.start()
                self.clients.append(new_thread)
            except KeyboardInterrupt:
                break
            except TimeoutError:
                if self.shutdown:
                    self.socket.shutdown(socket.SHUT_RDWR)
                    self.socket.close()
            except BaseException as e:
                self.logger.warning(str(e))

        if self.ssl:
            self.logger.info("SSL server has shut down")
        else:
            self.logger.info("TCP server has shut down")

    def launch_tcp_server(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('0.0.0.0', self.port))
        s.listen(1)

        return s

    def launch_ssl_server(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0) as sock:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            context = self.get_ssl_context()

            sconn = context.wrap_socket(sock, server_side=True)
            sconn.bind(('0.0.0.0', self.port))
            sconn.listen(0)

            return sconn

    def stop(self):
        if self.ssl:
            self.logger.warning("Shutting down SSL server")
        else:
            self.logger.warning("Shutting down TCP server")

        self.shutdown = True
        for client in self.clients:
            self.logger.debug('Attempting to stop client {}'.format(client.address))
            client.stop()

    def get_ssl_context(self):
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

        with self.app_context:
            context.load_cert_chain(
                os.path.join(self.app_context.app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver", "opentakserver.pem"),
                os.path.join(self.app_context.app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver", "opentakserver.nopass.key"))

            context.verify_mode = self.app_context.app.config.get("OTS_SSL_VERIFICATION_MODE")
            context.load_verify_locations(cafile=os.path.join(self.app_context.app.config.get("OTS_CA_FOLDER"), 'ca.pem'))

            return context



from dataclasses import dataclass
from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class Certificate(db.Model):
    __tablename__ = 'certificates'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    common_name: Mapped[str] = mapped_column(String)
    eud_uid: Mapped[str] = mapped_column(Integer, ForeignKey("euds.uid"), nullable=True)
    data_package_id: Mapped[int] = mapped_column(Integer, ForeignKey("data_packages.id"), nullable=True)
    callsign: Mapped[str] = mapped_column(String, nullable=True)
    username: Mapped[str] = mapped_column(String, ForeignKey("user.username"), nullable=True)
    expiration_date: Mapped[datetime] = mapped_column(DateTime)
    server_address: Mapped[str] = mapped_column(String)
    server_port: Mapped[int] = mapped_column(Integer)
    truststore_filename: Mapped[str] = mapped_column(String)
    user_cert_filename: Mapped[str] = mapped_column(String)
    csr: Mapped[str] = mapped_column(String, nullable=True)
    cert_password: Mapped[str] = mapped_column(String)
    user = relationship("User", back_populates="certificate", uselist=False)
    eud = relationship("EUD", back_populates="certificate", uselist=False)
    data_package = relationship("DataPackage", back_populates="certificate", uselist=False)

    def serialize(self):
        return {
            'callsign': self.callsign,
            'expiration_date': self.expiration_date,
            'server_address': self.server_address,
            'server_port': self.server_port,
            'truststore_filename': self.truststore_filename,
            'user_cert_filename': self.user_cert_filename,
            'cert_password': self.cert_password,
            'eud_uid': self.eud_uid
        }

    def to_json(self):
        return {
            'callsign': self.callsign,
            'expiration_date': iso8601_string_from_datetime(self.expiration_date),
            'server_address': self.server_address,
            'server_port': self.server_port,
            'truststore_filename': self.truststore_filename,
            'user_cert_filename': self.user_cert_filename,
            'data_package_filename': self.data_package.filename if self.data_package else None,
            'data_package_hash': self.data_package.hash if self.data_package else None,
            'eud_uid': self.eud_uid
        }



import secrets
import string
import random

import pyotp
from pathlib import Path
import os

import opentakserver


class DefaultConfig:
    SECRET_KEY = secrets.token_hex()
    DEBUG = False

    OTS_DATA_FOLDER = os.path.join(Path.home(), 'ots')
    OTS_LISTENER_ADDRESS = "127.0.0.1"
    OTS_LISTENER_PORT = 8081  # OTS will listen for HTTP requests on this port. Nginx will listen on ports 80, 443,
                              # 8080, 8443, and 8446 and proxy requests to OTS_LISTENER_PORT
    OTS_MARTI_HTTP_PORT = 8080
    OTS_MARTI_HTTPS_PORT = 8443
    OTS_ENABLE_TCP_STREAMING_PORT = True
    OTS_TCP_STREAMING_PORT = 8088
    OTS_SSL_STREAMING_PORT = 8089
    OTS_BACKUP_COUNT = 7
    OTS_RABBITMQ_SERVER_ADDRESS = "127.0.0.1"
    OTS_MEDIAMTX_API_ADDRESS = "http://localhost:9997"
    OTS_MEDIAMTX_TOKEN = str(secrets.SystemRandom().getrandbits(128))
    OTS_SSL_VERIFICATION_MODE = 2  # Equivalent to ssl.CERT_REQUIRED. https://docs.python.org/3/library/ssl.html#ssl.SSLContext.verify_mode
    OTS_NODE_ID = ''.join(random.choices(string.ascii_lowercase + string.digits, k=64))
    OTS_CA_NAME = 'OpenTAKServer-CA'
    OTS_CA_FOLDER = os.path.join(OTS_DATA_FOLDER, 'ca')
    OTS_CA_PASSWORD = 'atakatak'
    OTS_CA_EXPIRATION_TIME = 3650  # In days, defaults to 10 years
    OTS_CA_COUNTRY = 'WW'
    OTS_CA_STATE = 'XX'
    OTS_CA_CITY = 'YY'
    OTS_CA_ORGANIZATION = 'ZZ'
    OTS_CA_ORGANIZATIONAL_UNIT = 'OpenTAKServer'
    OTS_CA_SUBJECT = '/C={}/ST={}/L={}/O={}/OU={}'.format(OTS_CA_COUNTRY, OTS_CA_STATE, OTS_CA_CITY,
                                                          OTS_CA_ORGANIZATION, OTS_CA_ORGANIZATIONAL_UNIT)
    OTS_AIRPLANES_LIVE_LAT = 40.744213
    OTS_AIRPLANES_LIVE_LON = -73.986939
    OTS_AIRPLANES_LIVE_RADIUS = 10

    OTS_ENABLE_MUMBLE_AUTHENTICATION = False

    # Gmail settings
    OTS_ENABLE_EMAIL = False
    OTS_EMAIL_DOMAIN_WHITELIST = []
    OTS_EMAIL_DOMAIN_BLACKLIST = []
    OTS_EMAIL_TLD_WHITELIST = []
    OTS_EMAIL_TLD_BLACKLIST = []
    MAIL_SERVER = 'smtp.gmail.com'
    MAIL_PORT = 587
    MAIL_USE_SSL = False
    MAIL_USE_TLS = True
    MAIL_DEBUG = False
    MAIL_DEFAULT_SENDER = None
    MAIL_MAX_EMAILS = None
    MAIL_SUPPRESS_SEND = False
    MAIL_ASCII_ATTACHMENTS = False
    MAIL_USERNAME = None
    MAIL_PASSWORD = None

    # flask-sqlalchemy
    SQLALCHEMY_DATABASE_URI = "sqlite:////{}".format(os.path.join(OTS_DATA_FOLDER, 'ots.db'))
    SQLALCHEMY_ECHO = False
    SQLALCHEMY_ENGINE_OPTIONS = {"pool_pre_ping": True}
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_RECORD_QUERIES = False

    ALLOWED_EXTENSIONS = {'zip', 'xml'}

    UPLOAD_FOLDER = os.path.join(OTS_DATA_FOLDER, 'uploads')
    if not os.path.exists(UPLOAD_FOLDER):
        os.makedirs(UPLOAD_FOLDER)

    # Flask-Security-Too
    SECURITY_PASSWORD_SALT = str(secrets.SystemRandom().getrandbits(128))
    REMEMBER_COOKIE_SAMESITE = "strict"
    SESSION_COOKIE_SAMESITE = "strict"
    SECURITY_USERNAME_ENABLE = True
    SECURITY_USERNAME_REQUIRED = True
    SECURITY_TRACKABLE = True
    SECURITY_CSRF_COOKIE_NAME = "XSRF-TOKEN"
    WTF_CSRF_TIME_LIMIT = None
    SECURITY_CSRF_IGNORE_UNAUTH_ENDPOINTS = True
    WTF_CSRF_CHECK_DEFAULT = False
    SECURITY_RETURN_GENERIC_RESPONSES = True
    SECURITY_URL_PREFIX = "/api"
    SECURITY_CHANGEABLE = True
    SECURITY_CHANGE_URL = "/password/change"
    SECURITY_RESET_URL = "/password/reset"
    SECURITY_PASSWORD_LENGTH_MIN = 8
    SECURITY_REGISTERABLE = OTS_ENABLE_EMAIL
    SECURITY_CONFIRMABLE = OTS_ENABLE_EMAIL
    SECURITY_RECOVERABLE = OTS_ENABLE_EMAIL
    SECURITY_TWO_FACTOR = True
    SECURITY_TOTP_SECRETS = {1: pyotp.random_base32()}
    SECURITY_TOTP_ISSUER = "OpenTAKServer"
    SECURITY_TWO_FACTOR_ENABLED_METHODS = ["authenticator", "email"]
    SECURITY_TWO_FACTOR_RESCUE_MAIL = MAIL_USERNAME
    SECURITY_TWO_FACTOR_ALWAYS_VALIDATE = False
    SECURITY_LOGIN_WITHOUT_CONFIRMATION = True
    SECURITY_POST_CONFIRM_VIEW = "/login"
    SECURITY_REDIRECT_BEHAVIOR = 'spa'
    SECURITY_RESET_VIEW = '/reset'

    SCHEDULER_API_ENABLED = False
    #SCHEDULER_JOBSTORES = {'default': SQLAlchemyJobStore(url=SQLALCHEMY_DATABASE_URI)}



from jinja2 import Template

ca_config = """default_crl_days= 730                   # how long before next CRL

[ ca ]
        default_ca      = CA_default            # The default ca section

[ CA_default ]
        dir             = .                     # Where everything is kept
        certs           = $dir          # Where the issued certs are kept
        crl_dir         = $dir/crl              # Where the issued crl are kept
        database        = $dir/crl_index.txt    # database index file.
        default_md      = default               # use public key default MD

[ req ]
        default_bits           = 2048
        default_keyfile        = ca.pem
        distinguished_name     = req_distinguished_name
        x509_extensions        = v3_ca

[ req_distinguished_name ]
        countryName_min                = 2
        countryName_max                = 2
        commonName_max                 = 64

[ v3_ca ]
#basicConstraints=critical,CA:TRUE, pathlen:2
basicConstraints=critical,CA:TRUE
keyUsage=critical, cRLSign, keyCertSign
#nameConstraints=critical,permitted;DNS:.bbn.com # this allows you to restrict a CA to only issue server certs for a particular domain

[ client ]
basicConstraints=critical,CA:FALSE
keyUsage=critical, digitalSignature, keyEncipherment
extendedKeyUsage = critical, clientAuth
#extendedKeyUsage = critical, clientAuth, challengePassword
#authorityInfoAccess = OCSP;URI: http://localhost:4444

[ server ]
basicConstraints=critical,CA:FALSE
keyUsage=critical, digitalSignature, keyEncipherment
extendedKeyUsage = critical, clientAuth, serverAuth
#authorityInfoAccess = OCSP;URI: http://localhost:4444"""

server_config = Template("""default_crl_days= 730                   # how long before next CRL

[ ca ]
        default_ca      = CA_default            # The default ca section

[ CA_default ]
        dir             = .                     # Where everything is kept
        certs           = $dir          # Where the issued certs are kept
        crl_dir         = $dir/crl              # Where the issued crl are kept
        database        = $dir/crl_index.txt    # database index file.
        default_md      = default               # use public key default MD

[ req ]
        default_bits           = 2048
        default_keyfile        = ca.pem
        distinguished_name     = req_distinguished_name
        x509_extensions        = v3_ca

[ req_distinguished_name ]
        countryName_min                = 2
        countryName_max                = 2
        commonName_max                 = 64

[ v3_ca ]
#basicConstraints=critical,CA:TRUE, pathlen:2
basicConstraints=critical,CA:TRUE
keyUsage=critical, cRLSign, keyCertSign
#nameConstraints=critical,permitted;DNS:.bbn.com # this allows you to restrict a CA to only issue server certs for a particular domain

[ client ]
basicConstraints=critical,CA:FALSE
keyUsage=critical, digitalSignature, keyEncipherment
extendedKeyUsage = critical, clientAuth
#extendedKeyUsage = critical, clientAuth, challengePassword
#authorityInfoAccess = OCSP;URI: http://localhost:4444

[ server ]
basicConstraints=critical,CA:FALSE
keyUsage=critical, digitalSignature, keyEncipherment
extendedKeyUsage = critical, clientAuth, serverAuth
#authorityInfoAccess = OCSP;URI: http://localhost:4444

subjectAltName = @alt_names
[alt_names]
{{ alt_name_field }} = {{ common_name }}
""")


from dataclasses import dataclass

from flask_security.models import fsqla_v3 as fsqla
from opentakserver.extensions import db, logger


@dataclass
class Role(db.Model, fsqla.FsRoleMixin):

    def serialize(self):
        return {
            'name': self.name,
            'description': self.description,
            'permissions': self.permissions,
            'update_timestamp': self.update_datetime,
        }


import os
import threading
from threading import Timer
from eventlet.green import threading

import Ice

from opentakserver.mumble.mumble_authenticator import MumbleAuthenticator

# Load up Murmur slice file into Ice
Ice.loadSlice('', ['-I' + Ice.getSliceDir(), os.path.join(os.path.dirname(os.path.realpath(__file__)), 'Murmur.ice')])
import Murmur


class MumbleIceDaemon(threading.Thread):
    def __init__(self, app, logger):
        super().__init__()
        self.app = app
        self.logger = logger
        self.logger.info("mumble daemon init")
        self.daemon = True

    def run(self):
        # Configure Ice properties
        props = Ice.createProperties()
        props.setProperty("Ice.ImplicitContext", "Shared")
        props.setProperty('Ice.Default.EncodingVersion', '1.0')
        props.setProperty('Ice.Default.InvocationTimeout', str(30 * 1000))
        props.setProperty('Ice.MessageSizeMax', str(1024))
        idata = Ice.InitializationData()
        idata.properties = props

        # Create Ice connection
        ice = Ice.initialize(idata)
        proxy = ice.stringToProxy('Meta:tcp -h 127.0.0.1 -p 6502')
        secret = ''
        if secret != '':
            ice.getImplicitContext().put("secret", secret)
        try:
            meta = Murmur.MetaPrx.checkedCast(proxy)
        except Ice.ConnectionRefusedException:
            self.logger.error("Failed to connect to the mumble ice server")
            return

        mumble_ice_app = MumbleIceApp(self.app, self.logger, ice)
        mumble_ice_app.run()


class MumbleIceApp(Ice.Application):
    def __init__(self, app, logger, ice):
        super().__init__()
        self.app = app
        self.logger = logger
        self.ice = ice
        self.meta = None
        self.metacb = None
        self.connected = False
        self.failed_watch = False
        self.watchdog = None
        self.auth = None

    def run(self, *args):
        if not self.initialize_ice_connection():
            self.logger.error("Mumble server connection failed")
            return 1

        self.check_connection()

        self.watchdog.cancel()

        if self.interrupted():
            self.logger.warning('Caught interrupt, shutting down')

        return 0

    def initialize_ice_connection(self):
        """
        Establishes the two-way Ice connection and adds the authenticator to the
        configured servers
        """

        # if False and 'ice_secret':
        #     self.ice.getImplicitContext().put("secret", "some_secret")

        self.logger.debug('Connecting to Ice server ({}:{})'.format('127.0.0.1', 6502))
        base = self.ice.stringToProxy('Meta:tcp -h {} -p {}'.format('127.0.0.1', 6502))
        self.meta = Murmur.MetaPrx.uncheckedCast(base)

        adapter = self.ice.createObjectAdapterWithEndpoints('Callback.Client', 'tcp -h 127.0.0.1')
        adapter.activate()

        metacbprx = adapter.addWithUUID(MetaCallback(self))
        self.metacb = Murmur.MetaCallbackPrx.uncheckedCast(metacbprx)

        authprx = adapter.addWithUUID(MumbleAuthenticator(self.app, self.logger, self.ice))
        self.auth = Murmur.ServerUpdatingAuthenticatorPrx.uncheckedCast(authprx)

        return self.attach_callbacks()

    def attach_callbacks(self):
        """
        Attaches all callbacks for meta and authenticators
        """

        try:
            self.logger.debug('Attaching meta callback')

            self.meta.addCallback(self.metacb)

            for server in self.meta.getBootedServers():
                self.logger.debug('Setting mumble authenticator for virtual server {}'.format(server.id()))
                server.setAuthenticator(self.auth)

        except (Murmur.InvalidSecretException, Ice.UnknownUserException, Ice.ConnectionRefusedException) as e:
            if isinstance(e, Ice.ConnectionRefusedException):
                self.logger.warning('Server refused connection')
            elif isinstance(e, Murmur.InvalidSecretException) or \
                    isinstance(e, Ice.UnknownUserException) and (e.unknown == 'Murmur::InvalidSecretException'):
                self.logger.error('Invalid ice secret')
            else:
                # We do not actually want to handle this one, re-raise it
                raise e

            self.connected = False
            return False

        self.connected = True
        return True

    def check_connection(self):
        """
        Tries reapplies all callbacks to make sure the authenticator
        survives server restarts and disconnects.
        """

        try:
            self.attach_callbacks()
        except Ice.Exception as e:
            self.logger.warning('{}: Failed connection check, will retry in next watchdog run ({}s)'.format(e, 10))

        # Renew the timer
        self.watchdog = Timer(10, self.check_connection)
        self.watchdog.start()


class MetaCallback(Murmur.MetaCallback):
    def __init__(self, authenticator):
        Murmur.MetaCallback.__init__(self)
        self.authenticator = authenticator

    def started(self, server, current=None):
        """
        This function is called when a virtual server is started
        and makes sure an authenticator gets attached if needed.
        """
        self.authenticator.logger.info('Setting authenticator for virtual server {}'.format(server.id()))
        try:
            server.setAuthenticator(self.authenticator.auth)
        # Apparently this server was restarted without us noticing
        except (Murmur.InvalidSecretException, Ice.UnknownUserException) as e:
            if hasattr(e, "unknown") and e.unknown != "Murmur::InvalidSecretException":
                # Special handling for Murmur 1.2.2 servers with invalid slice files
                raise e

            return

    def stopped(self, server, current=None):
        """
        This function is called when a virtual server is stopped
        """
        if self.authenticator.connected:
            # Only try to output the server id if we think we are still connected to prevent
            # flooding of our thread pool
            try:
                self.authenticator.logger.info('Authenticated virtual server {} got stopped'.format(server.id()))
                return
            except Ice.ConnectionRefusedException:
                self.authenticator.connected = False

        self.authenticator.logger.info('Server shutdown stopped a virtual server')



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class Team(db.Model):
    __tablename__ = "teams"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String, unique=True)
    chatroom_id: Mapped[int] = mapped_column(Integer, ForeignKey("chatrooms.id"), nullable=True)
    euds = relationship("EUD", back_populates="team")
    chatroom = relationship("Chatroom", back_populates="team")

    colors = {'Cyan': '#00FFFF', 'White': '#000000', 'Yellow': '#FFFF00', 'Orange': '#FFA500', 'Magenta': '#FF00FF',
              'Red': '#FF0000', 'Maroon': '#800000', 'Purple': '#800080', 'Dark Blue': '#00008B', 'Blue': '#0000FF',
              'Teal': '#008080', 'Green': '#00FF00', 'Dark Green': '#228B22', 'Brown': '#964B00'}

    def get_team_color(self):
        return self.colors[self.name]

    def serialize(self):
        return {
            'name': self.name,
        }

    def to_json(self):
        return {
            'name': self.name,
            'chatroom': self.chatroom.to_json() if self.chatroom else None,
            'euds': [eud.to_json() for eud in self.euds] if self.euds else None,
            'color': self.colors[self.name]
        }



from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Boolean, Float, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.Icon import Icon


class CasEvac(db.Model):
    __tablename__ = 'casevac'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    sender_uid: Mapped[str] = mapped_column(String, ForeignKey("euds.uid"))
    uid: Mapped[str] = mapped_column(String, unique=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)

    # The following are taken from CoT attributes generated from ATAK
    ambulatory: Mapped[int] = mapped_column(Integer, nullable=True)
    casevac: Mapped[bool] = mapped_column(Boolean, nullable=True)
    child: Mapped[int] = mapped_column(Integer, nullable=True)
    enemy: Mapped[str] = mapped_column(String, nullable=True)
    epw: Mapped[int] = mapped_column(Integer, nullable=True)
    equipment_detail: Mapped[str] = mapped_column(String, nullable=True)
    equipment_none: Mapped[bool] = mapped_column(Boolean, nullable=True)
    equipment_other: Mapped[bool] = mapped_column(Boolean, nullable=True)
    extraction_equipment: Mapped[bool] = mapped_column(Boolean, nullable=True)
    freq: Mapped[float] = mapped_column(Float, nullable=True)
    friendlies: Mapped[str] = mapped_column(String, nullable=True)
    hlz_marking: Mapped[int] = mapped_column(Integer, nullable=True)
    hlz_remarks: Mapped[str] = mapped_column(String, nullable=True)
    hoist: Mapped[bool] = mapped_column(Boolean, nullable=True)
    litter: Mapped[int] = mapped_column(Integer, nullable=True)
    marked_by: Mapped[str] = mapped_column(String, nullable=True)
    medline_remarks: Mapped[str] = mapped_column(String, nullable=True)
    nonus_civilian: Mapped[int] = mapped_column(Integer, nullable=True)
    nonus_military: Mapped[int] = mapped_column(Integer, nullable=True)
    obstacles: Mapped[str] = mapped_column(String, nullable=True)
    priority: Mapped[int] = mapped_column(Integer, nullable=True)
    routine: Mapped[int] = mapped_column(Integer, nullable=True)
    security: Mapped[int] = mapped_column(Integer, nullable=True)
    terrain_loose: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_other: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_other_detail: Mapped[bool] = mapped_column(String, nullable=True)
    terrain_detail: Mapped[str] = mapped_column(String, nullable=True)
    terrain_none: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_rough: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_slope: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_slope_dir: Mapped[str] = mapped_column(String, nullable=True)
    title: Mapped[str] = mapped_column(String)
    urgent: Mapped[int] = mapped_column(Integer, nullable=True)
    us_civilian: Mapped[int] = mapped_column(Integer, nullable=True)
    us_military: Mapped[int] = mapped_column(Integer, nullable=True)
    ventilator: Mapped[bool] = mapped_column(Boolean, nullable=True)
    winds_are_from: Mapped[str] = mapped_column(String, nullable=True)
    zone_prot_selection: Mapped[int] = mapped_column(Integer, nullable=True)
    point_id: Mapped[int] = mapped_column(Integer, ForeignKey("points.id"), nullable=True)
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id"), nullable=True)
    point = relationship("Point", back_populates="casevac")
    cot = relationship("CoT", back_populates="casevac")
    zmist = relationship("ZMIST", back_populates="casevac", uselist=False)
    eud = relationship("EUD", back_populates="casevacs")

    def serialize(self):
        return {
            'sender_uid': self.sender_uid,
            'uid': self.uid,
            'timestamp': self.timestamp,
            'ambulatory': self.ambulatory,
            'casevac': self.casevac,
            'child': self.child,
            'enemy': self.enemy,
            'epw': self.epw,
            'equipment_detail': self.equipment_detail,
            'equipment_none': self.equipment_none,
            'equipment_other': self.equipment_other,
            'extraction_equipment': self.extraction_equipment,
            'freq': self.freq,
            'friendlies': self.friendlies,
            'hlz_marking': self.hlz_marking,
            'hlz_remarks': self.hlz_remarks,
            'hoist': self.hoist,
            'litter': self.litter,
            'marked_by': self.marked_by,
            'medline_remarks': self.medline_remarks,
            'nonus_civilian': self.nonus_civilian,
            'nonus_military': self.nonus_military,
            'obstacles': self.obstacles,
            'priority': self.priority,
            'routine': self.routine,
            'security': self.security,
            'terrain_loose': self.terrain_loose,
            'terrain_other': self.terrain_other,
            'terrain_other_detail': self.terrain_other_detail,
            'terrain_detail': self.terrain_detail,
            'terrain_none': self.terrain_none,
            'terrain_rough': self.terrain_rough,
            'terrain_slope': self.terrain_slope,
            'terrain_slope_dir': self.terrain_slope_dir,
            'title': self.title,
            'urgent': self.urgent,
            'us_civilian': self.us_civilian,
            'us_military': self.us_military,
            'ventilator': self.ventilator,
            'winds_are_from': self.winds_are_from,
            'zone_prot_selection': self.zone_prot_selection,
            'eud': self.eud,
        }

    def to_json(self):
        icon = db.session.execute(db.session.query(Icon).filter(Icon.filename == 'red_crs.png')).first()[0]
        return {
            'sender_uid': self.sender_uid,
            'uid': self.uid,
            'timestamp': iso8601_string_from_datetime(self.timestamp),
            'ambulatory': self.ambulatory,
            'casevac': self.casevac,
            'child': self.child,
            'enemy': self.enemy,
            'epw': self.epw,
            'equipment_detail': self.equipment_detail,
            'equipment_none': self.equipment_none,
            'equipment_other': self.equipment_other,
            'extraction_equipment': self.extraction_equipment,
            'freq': self.freq,
            'friendlies': self.friendlies,
            'hlz_marking': self.hlz_marking,
            'hlz_remarks': self.hlz_remarks,
            'hoist': self.hoist,
            'litter': self.litter,
            'marked_by': self.marked_by,
            'medline_remarks': self.medline_remarks,
            'nonus_civilian': self.nonus_civilian,
            'nonus_military': self.nonus_military,
            'obstacles': self.obstacles,
            'priority': self.priority,
            'routine': self.routine,
            'security': self.security,
            'terrain_loose': self.terrain_loose,
            'terrain_other': self.terrain_other,
            'terrain_other_detail': self.terrain_other_detail,
            'terrain_detail': self.terrain_detail,
            'terrain_none': self.terrain_none,
            'terrain_rough': self.terrain_rough,
            'terrain_slope': self.terrain_slope,
            'terrain_slope_dir': self.terrain_slope_dir,
            'title': self.title,
            'urgent': self.urgent,
            'us_civilian': self.us_civilian,
            'us_military': self.us_military,
            'ventilator': self.ventilator,
            'winds_are_from': self.winds_are_from,
            'zone_prot_selection': self.zone_prot_selection,
            'zmist': self.zmist.serialize() if self.zmist else None,
            'eud': self.eud,
            'point': self.point.to_json() if self.point else None,
            'icon': icon.to_json(),
            'start': iso8601_string_from_datetime(self.cot.start) if self.cot else None,
            'stale': iso8601_string_from_datetime(self.cot.stale) if self.cot else None,
        }



from dataclasses import dataclass
from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class DataPackage(db.Model):
    __tablename__ = 'data_packages'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    filename: Mapped[str] = mapped_column(String, unique=True)
    hash: Mapped[str] = mapped_column(String, unique=True)
    creator_uid: Mapped[str] = mapped_column(String, ForeignKey("euds.uid"))
    submission_time: Mapped[datetime] = mapped_column(DateTime)
    submission_user: Mapped[int] = mapped_column(Integer, ForeignKey("user.id"), nullable=True)
    keywords: Mapped[str] = mapped_column(String, nullable=True)
    mime_type: Mapped[str] = mapped_column(String)
    size: Mapped[int] = mapped_column(Integer)
    tool: Mapped[str] = mapped_column(String, nullable=True)
    expiration: Mapped[str] = mapped_column(String, nullable=True)
    eud: Mapped["EUD"] = relationship(back_populates="data_packages")
    certificate = relationship("Certificate", back_populates="data_package", uselist=False)
    user = relationship("User", back_populates="data_packages")

    def serialize(self):
        return {
            'filename': self.filename,
            'hash': self.hash,
            'creator_uid': self.creator_uid,
            'submission_time': self.submission_time,
            'submission_user': self.submission_user,
            'keywords': self.keywords,
            'mime_type': self.mime_type,
            'size': self.size,
            'tool': self.tool,
            'expiration': self.expiration,
        }

    def to_json(self, include_eud=True):
        return {
            'filename': self.filename,
            'hash': self.hash,
            'creator_uid': self.creator_uid,
            'submission_time': iso8601_string_from_datetime(self.submission_time),
            'submission_user': self.user.username if self.user else None,
            'keywords': self.keywords,
            'mime_type': self.mime_type,
            'size': self.size,
            'tool': self.tool,
            'expiration': self.expiration,
            'eud': self.eud.to_json(False) if include_eud and self.eud else None
        }


import os

import Ice

from flask_security import verify_password

# Load up Murmur slice file into Ice
Ice.loadSlice('', ['-I' + Ice.getSliceDir(), os.path.join(os.path.dirname(os.path.realpath(__file__)), 'Murmur.ice')])
import Murmur


class MumbleAuthenticator(Murmur.ServerUpdatingAuthenticator):
    texture_cache = {}

    def __init__(self, app, logger, ice):
        Murmur.ServerUpdatingAuthenticator.__init__(self)
        self.app = app
        self.logger = logger
        self.ice = ice

    def authenticate(self, username, password, certlist, certhash, strong, current=None):
        """
        This function is called to authenticate a user
        """
        if username == 'SuperUser':
            return (-2, None, None)

        self.logger.info("Mumble auth request for {}".format(username))

        with self.app.app_context():
            user = self.app.security.datastore.find_user(username=username)
            if not user:
                self.logger.warning("Mumble auth: User {} not found".format(username))
                return (-1, None, None)
            elif not user.active:
                self.logger.warning("Mumble auth: User {} is deactivated".format(username))
                return (-1, None, None)

            if verify_password(password, user.password):
                self.logger.info("Mumble auth: {} has been authenticated".format(username))
                return (user.id, user.username, None)

            self.logger.warning("Mumble auth: Bad password for {}".format(username))
            return (-1, None, None)

    def idToTexture(self, id, current=None):
        return

    def getInfo(self, id, current=None):
        """
        Gets called to fetch user specific information
        """

        # We do not expose any additional information so always fall through
        return (False, None)

    def nameToId(self, name, current=None):
        """
        Gets called to get the id for a given username
        """
        pass

    def idToName(self, id, current=None):
        """
        Gets called to get the username for a given id
        """
        pass

    def idToTexture(self, id, current=None):
        """
        Gets called to get the corresponding texture for a user
        """
        # seems like it pulled a user's avatar from a phpbb DB

    def registerUser(self, name, current=None):
        """
        Gets called when the server is asked to register a user.
        """
        pass

    def unregisterUser(self, id, current=None):
        """
        Gets called when the server is asked to unregister a user.
        """
        pass

    def getRegisteredUsers(self, filter, current=None):
        """
        Returns a list of usernames in the phpBB3 database which contain
        filter as a substring.
        """
        pass

    def setInfo(self, id, info, current=None):
        """
        Gets called when the server is supposed to save additional information
        about a user to his database
        """
        pass

    def setTexture(self, id, texture, current=None):
        """
        Gets called when the server is asked to update the user texture of a user
        """
        pass


from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Boolean, Float
from sqlalchemy.orm import Mapped, mapped_column, relationship


class ZMIST(db.Model):
    __tablename__ = 'zmist'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    i: Mapped[str] = mapped_column(String, nullable=True)  # injury_sustained
    m: Mapped[str] = mapped_column(String, nullable=True)  # mechanism_of_injury
    s: Mapped[str] = mapped_column(String, nullable=True)  # symptoms_and_signs
    t: Mapped[str] = mapped_column(String, nullable=True)  # treatment_given
    title: Mapped[str] = mapped_column(String, nullable=True)
    z: Mapped[int] = mapped_column(Integer, nullable=True)  # zap_number
    casevac_uid: Mapped[str] = mapped_column(String, ForeignKey("casevac.uid"))
    casevac = relationship("CasEvac", back_populates="zmist")

    def serialize(self):
        return {
            'i': self.i,
            'm': self.m,
            's': self.s,
            't': self.t,
            'title': self.title,
            'z': self.z
        }

    def to_json(self):
        return self.serialize()



from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship


class GeoChat(db.Model):
    __tablename__ = 'geochat'

    uid: Mapped[str] = mapped_column(String, primary_key=True)
    chatroom_id: Mapped[str] = mapped_column(String, ForeignKey("chatrooms.id"))
    sender_uid: Mapped[str] = mapped_column(String, ForeignKey("euds.uid"))
    remarks: Mapped[str] = mapped_column(String)
    timestamp: Mapped[datetime] = mapped_column(DateTime)
    point_id: Mapped[int] = mapped_column(Integer, ForeignKey("points.id"))
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id"))
    point = relationship("Point", back_populates="geochat")
    cot = relationship("CoT", back_populates="geochat")
    chatroom = relationship("Chatroom", back_populates="geochats")
    eud = relationship("EUD", back_populates="geochats")

    def serialize(self):
        return {
            'uid': self.uid,
            'sender_uid': self.sender_uid,
            'remarks': self.remarks,
            'timestamp': self.timestamp,
        }

    def to_json(self):
        return {
            'uid': self.uid,
            'sender_uid': self.sender_uid,
            'remarks': self.remarks,
            'timestamp': self.timestamp,
            'point': self.point.to_json() or None,
        }



import base64
import hashlib
import json
import os
import datetime
import time
import traceback
import uuid
from urllib.parse import urlparse

import jwt

from xml.etree.ElementTree import Element, tostring, fromstring, SubElement

import bleach
import sqlalchemy
from OpenSSL import crypto
from bs4 import BeautifulSoup
from flask import current_app as app, request, Blueprint, send_from_directory, jsonify
from flask_security import verify_password, current_user
from opentakserver.extensions import logger, db
from opentakserver.forms.MediaMTXPathConfig import MediaMTXPathConfig
from opentakserver import __version__ as version

from opentakserver.models.EUD import EUD
from opentakserver.models.DataPackage import DataPackage
from werkzeug.utils import secure_filename

from opentakserver.models.VideoStream import VideoStream

from opentakserver.certificate_authority import CertificateAuthority

from opentakserver.models.Certificate import Certificate

marti_blueprint = Blueprint('marti_blueprint', __name__)


# flask-security's http_auth_required() decorator will deny access because ATAK doesn't do CSRF,
# so we handle basic auth ourselves
def basic_auth(credentials):
    try:
        username, password = base64.b64decode(credentials.split(" ")[-1].encode('utf-8')).decode('utf-8').split(":")
        username = bleach.clean(username)
        password = bleach.clean(password)
        user = app.security.datastore.find_user(username=username)
        return user and verify_password(password, user.password)
    except BaseException as e:
        logger.error("Failed to verify credentials: {}".format(e))
        return False


@marti_blueprint.route('/Marti/api/clientEndPoints', methods=['GET'])
def client_end_points():
    euds = db.session.execute(db.select(EUD)).scalars()
    return_value = {'version': 3, "type": "com.bbn.marti.remote.ClientEndpoint", 'data': [],
                    'nodeId': app.config.get("OTS_NODE_ID")}
    for eud in euds:
        return_value['data'].append({
            'callsign': eud.callsign,
            'uid': eud.uid,
            'username': current_user.username if current_user.is_authenticated else 'anonymous',
            'lastEventTime': eud.last_event_time,
            'lastStatus': eud.last_status
        })

    return return_value, 200, {'Content-Type': 'application/json'}


# require basic auth
@marti_blueprint.route('/Marti/api/tls/config')
def tls_config():
    if not basic_auth(request.headers.get('Authorization')):
        return '', 401

    root_element = Element('ns2:certificateConfig')
    root_element.set('xmlns', "http://bbn.com/marti/xml/config")
    root_element.set('xmlns:ns2', "com.bbn.marti.config")

    name_entries = SubElement(root_element, "nameEntries")
    first_name_entry = SubElement(name_entries, "nameEntry")
    first_name_entry.set('name', 'O')
    first_name_entry.set('value', 'Test Organization Name')

    second_name_entry = SubElement(name_entries, "nameEntry")
    second_name_entry.set('name', 'OU')
    second_name_entry.set('value', 'Test Organization Unit Name')

    return tostring(root_element), 200, {'Content-Type': 'application/xml'}


@marti_blueprint.route('/Marti/api/tls/profile/enrollment')
def enrollment():
    if not basic_auth(request.headers.get('Authorization')):
        return '', 401
    logger.info("Enrollment request from {}".format(request.args.get('clientUid')))
    return '', 204


@marti_blueprint.route('/Marti/api/tls/signClient/', methods=['POST'])
def sign_csr():
    if not basic_auth(request.headers.get('Authorization')):
        return '', 401
    return '', 200


@marti_blueprint.route('/Marti/api/tls/signClient/v2', methods=['POST'])
def sign_csr_v2():
    if not basic_auth(request.headers.get('Authorization')):
        return '', 401

    try:
        uid = request.args.get("clientUid")

        if "iTAK" not in request.user_agent.string:
            csr = '-----BEGIN CERTIFICATE REQUEST-----\n' + request.data.decode(
                'utf-8') + '-----END CERTIFICATE REQUEST-----'
        else:
            csr = request.data.decode('utf-8')

        x509 = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr.encode())
        common_name = x509.get_subject().CN
        logger.debug("Attempting to sign CSR for {}".format(common_name))

        cert_authority = CertificateAuthority(logger, app)

        signed_csr = cert_authority.sign_csr(csr.encode(), common_name, False).decode("utf-8")
        signed_csr = signed_csr.replace("-----BEGIN CERTIFICATE-----\n", "")
        signed_csr = signed_csr.replace("\n-----END CERTIFICATE-----\n", "")

        f = open(os.path.join(app.config.get("OTS_CA_FOLDER"), "ca.pem"), 'r')
        cert = f.read()
        f.close()

        cert = cert.replace("-----BEGIN CERTIFICATE-----\n", "")
        cert = cert.replace("\n-----END CERTIFICATE-----\n", "")

        if "iTAK" in request.user_agent.string:
            response = {'signedCert': signed_csr, 'ca0': cert, 'ca1': cert}
        else:
            enrollment = Element('enrollment')
            signed_cert = SubElement(enrollment, 'signedCert')
            signed_cert.text = signed_csr
            ca = SubElement(enrollment, 'ca')
            ca.text = cert

            response = tostring(enrollment).decode('utf-8')
            response = '<?xml version="1.0" encoding="UTF-8"?>\n' + response

        username, password = base64.b64decode(
            request.headers.get("Authorization").split(" ")[-1].encode('utf-8')).decode(
            'utf-8').split(":")
        username = bleach.clean(username)
        user = app.security.datastore.find_user(username=username)

        try:
            eud = EUD()
            eud.uid = uid
            eud.user_id = user.id

            db.session.add(eud)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()
            eud = db.session.execute(db.session.query(EUD).filter_by(uid=uid)).first()[0]
            if user and not eud.user_id:
                eud.user_id = user.id
                db.session.add(eud)
                db.session.commit()

        try:
            certificate = Certificate()
            certificate.common_name = common_name
            certificate.eud_uid = uid
            certificate.callsign = eud.callsign
            certificate.expiration_date = datetime.datetime.today() + datetime.timedelta(
                days=app.config.get("OTS_CA_EXPIRATION_TIME"))
            certificate.server_address = urlparse(request.url_root).hostname
            certificate.server_port = app.config.get("OTS_MARTI_HTTPS_PORT")
            certificate.truststore_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "truststore-root.p12")
            certificate.user_cert_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                                          common_name + ".pem")
            certificate.csr = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr")
            certificate.cert_password = app.config.get("OTS_CA_PASSWORD")

            db.session.add(certificate)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()
            certificate = db.session.execute(db.session.query(Certificate).filter_by(eud_uid=eud.uid)).scalar_one()
            certificate.common_name = common_name
            certificate.callsign = eud.callsign
            certificate.expiration_date = datetime.datetime.today() + datetime.timedelta(
                days=app.config.get("OTS_CA_EXPIRATION_TIME"))
            certificate.server_address = urlparse(request.url_root).hostname
            certificate.server_port = app.config.get("OTS_MARTI_HTTPS_PORT")
            certificate.truststore_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "truststore-root.p12")
            certificate.user_cert_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                                          common_name + ".pem")
            certificate.csr = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr")
            certificate.cert_password = app.config.get("OTS_CA_PASSWORD")

            db.session.commit()

        if "iTAK" in request.user_agent.string:
            return response, 200, {'Content-Type': 'text/plain', 'Content-Encoding': 'charset=UTF-8'}
        else:
            return response, 200, {'Content-Type': 'application/xml', 'Content-Encoding': 'charset=UTF-8'}
    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({'error': str(e)}), 500


@marti_blueprint.route('/Marti/api/version/config', methods=['GET'])
def marti_config():
    url = urlparse(request.url_root)

    return {"version": "3", "type": "ServerConfig",
            "data": {"version": version, "api": "3", "hostname": url.hostname},
            "nodeId": app.config.get("OTS_NODE_ID")}, 200, {'Content-Type': 'application/json'}


@marti_blueprint.route('/Marti/api/missions/citrap/subscription', methods=['PUT'])
def citrap_subscription():
    uid = bleach.clean(request.args.get('uid'))
    response = {
        'version': 3, 'type': 'com.bbn.marti.sync.model.MissionSubscription',
        'data': {

        }
    }
    return '', 201


@marti_blueprint.route('/Marti/api/missions/invitations')
def mission_invitations():
    uid = bleach.clean(request.args.get('clientUid'))
    response = {
        'version': 3, 'type': 'MissionInvitation', 'data': [], 'nodeId': app.config.get('OTS_NODE_ID')
    }

    return jsonify(response)


@marti_blueprint.route('/Marti/api/missions')
def missions():
    password_protected = request.args.get('passwordProtected')
    if password_protected:
        password_protected = bleach.clean(password_protected).lower() == 'true'

    default_role = request.args.get('defaultRole')
    if default_role:
        default_role = bleach.clean(default_role).lower() == 'true'

    response = {
        'version': 3, 'type': 'Mission', 'data': [], 'nodeId': app.config.get('OTS_NODE_ID')
    }

    return jsonify(response)


@marti_blueprint.route('/Marti/api/groups/all')
def groups():
    use_cache = bleach.clean(request.args.get('useCache'))  # bool

    response = {
        'version': 3, 'type': 'com.bbn.marti.remote.groups.Group', 'data': [{}], 'nodeId': app.config.get("OTS_NODE_ID")
    }

    return jsonify(response)


@marti_blueprint.route('/Marti/api/missions/all/invitations')
def all_invitations():
    clientUid = bleach.clean(request.args.get('clientUid'))
    return '', 200


@marti_blueprint.route('/Marti/api/missions/<mission_name>', methods=['GET', 'PUT'])
def put_mission(mission_name):
    if request.method == 'PUT':
        creator_uid = bleach.clean(request.args.get('creatorUid'))
        description = bleach.clean(request.args.get('description'))
        tool = bleach.clean(request.args.get('tool'))
        group = bleach.clean(request.args.get('group'))
        default_role = bleach.clean(request.args.get('defaultRole'))
        password = request.args.get('password')
        password_protected = False
        if password:
            password = bleach.clean(password)
            password_protected = True

        uid = str(uuid.uuid4())
        creation_time = int(time.time())
        creation_datetime = datetime.datetime.now()

        payload = {'jti': uid, 'iat': creation_time, 'sub': 'SUBSCRIPTION', 'iss': '',
                   'SUBSCRIPTION': uid, 'MISSION_NAME': mission_name}

        server_key = open(os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver",
                                       "opentakserver.nopass.key"), "rb")

        token = jwt.encode(payload, server_key.read(), algorithm="HS256")
        server_key.close()

        response = {
            'version': 3, 'type': 'Mission', 'data': [{
                'name': mission_name, 'description': description, 'chatRoom': '', 'baseLayer': '', 'path': '',
                'classification': '',
                'tool': tool, 'keywords': [], 'creatorUid': creator_uid, 'createTime': creation_datetime,
                'externalData': [],
                'feeds': [], 'mapLayers': [], 'defaultRole': {
                    'permissions': ["MISSION_WRITE", "MISSION_READ"], "type": "MISSION_SUBSCRIBER"},

                'ownerRole': {"permissions": ["MISSION_MANAGE_FEEDS", "MISSION_SET_PASSWORD", "MISSION_WRITE",
                                              "MISSION_MANAGE_LAYERS", "MISSION_UPDATE_GROUPS", "MISSION_SET_ROLE",
                                              "MISSION_READ", "MISSION_DELETE"], "type": "MISSION_OWNER"},
                'inviteOnly': False, 'expiration': -1, 'guid': '', 'uids': [], 'contents': [], 'token': token,
                'passwordProtected': password_protected, 'nodeId': app.config.get("OTS_NODE_ID")
            }]
        }

        return jsonify(response), 201
    elif request.method == 'GET':
        # Pull it from the DB

        return '', 200


@marti_blueprint.route('/Marti/api/missions/<mission_name>/keywords', methods=['PUT'])
def put_mission_keywords(mission_name):
    keywords = request.json()

    return '', 200


@marti_blueprint.route('/Marti/api/missions/<mission_name>/subscription', methods=['PUT'])
def mission_subscribe(mission_name):
    uid = bleach.clean(request.args.get("uid"))

    return '', 200


@marti_blueprint.route('/Marti/api/citrap')
def citrap():
    return jsonify([])


@marti_blueprint.route('/Marti/api/groups/groupCacheEnabled')
def group_cache_enabled():
    response = {
        'version': 3, 'type': 'java.lang.Boolean', 'data': False, 'nodeId': app.config.get('OTS_NODE_ID')
    }

    return jsonify(response)


@marti_blueprint.route('/Marti/sync/upload', methods=['POST'])
def itak_data_package_upload():
    if not request.content_length:
        return {'error': 'no file'}, 400, {'Content-Type': 'application/json'}
    elif request.content_type != 'application/x-zip-compressed':
        logger.error("Not a zip")
        return {'error': 'Please only upload zip files'}, 415, {'Content-Type': 'application/json'}

    file = request.data
    sha256 = hashlib.sha256()
    sha256.update(file)
    file_hash = sha256.hexdigest()
    logger.debug("got sha256 {}".format(file_hash))
    hash_filename = secure_filename(file_hash + '.zip')

    with open(os.path.join(app.config.get("UPLOAD_FOLDER"), hash_filename), "wb") as f:
        f.write(file)

    try:
        data_package = DataPackage()
        data_package.filename = request.args.get('name')
        data_package.hash = file_hash
        data_package.creator_uid = request.args.get('CreatorUid') if request.args.get('CreatorUid') else str(
            uuid.uuid4())
        data_package.submission_user = current_user.id if current_user.is_authenticated else None
        data_package.submission_time = datetime.datetime.now()
        data_package.mime_type = request.content_type
        data_package.size = os.path.getsize(os.path.join(app.config.get("UPLOAD_FOLDER"), hash_filename))
        db.session.add(data_package)
        db.session.commit()
    except sqlalchemy.exc.IntegrityError as e:
        db.session.rollback()
        logger.error("Failed to save data package: {}".format(e))
        return jsonify({'success': False, 'error': 'This data package has already been uploaded'}), 400

    return_value = {"UID": data_package.hash, "SubmissionDateTime": data_package.submission_time,
                    "Keywords": ["missionpackage"],
                    "MIMEType": data_package.mime_type, "SubmissionUser": "anonymous", "PrimaryKey": "1",
                    "Hash": data_package.hash, "CreatorUid": data_package.creator_uid, "Name": data_package.filename}

    return jsonify(return_value)


@marti_blueprint.route('/Marti/sync/missionupload', methods=['POST'])
def data_package_share():
    if not len(request.files):
        return {'error': 'no file'}, 400, {'Content-Type': 'application/json'}
    for file in request.files:
        file = request.files[file]

        if file.content_type != 'application/x-zip-compressed' and file.content_type != "application/zip-compressed" \
                and file.content_type != "application/zip" and not file.content_type.startswith("application/x-zip"):
            logger.error("Uploaded data package does not seem to be a zip file. The content type is {}".format(
                file.content_type))
            return {'error': "Uploaded data package does not seem to be a zip file. The content type is {}".format(
                file.content_type)}, 415, {'Content-Type': 'application/json'}

        if file:
            file_hash = request.args.get('hash')
            if not file_hash:
                sha256 = hashlib.sha256()
                sha256.update(file.stream.read())
                file.stream.seek(0)
                file_hash = sha256.hexdigest()
                logger.debug("got sha256 {}".format(file_hash))

            logger.debug("Got file: {} - {}".format(file.filename, file_hash))

            filename = secure_filename(file_hash + '.zip')
            file.save(os.path.join(app.config.get("UPLOAD_FOLDER"), filename))

            try:
                data_package = DataPackage()
                data_package.filename = file.filename
                data_package.hash = file_hash
                data_package.creator_uid = request.args.get('creatorUid') if request.args.get('creatorUid') else str(
                    uuid.uuid4())
                data_package.submission_user = current_user.id if current_user.is_authenticated else None
                data_package.submission_time = datetime.datetime.now()
                data_package.mime_type = file.mimetype
                data_package.size = os.path.getsize(os.path.join(app.config.get("UPLOAD_FOLDER"), filename))
                db.session.add(data_package)
                db.session.commit()
            except sqlalchemy.exc.IntegrityError as e:
                db.session.rollback()
                logger.error("Failed to save data package: {}".format(e))
                return jsonify({'success': False, 'error': 'This data package has already been uploaded'}), 400

            url = urlparse(request.url_root)
            return 'https://{}:{}/Marti/api/sync/metadata/{}/tool'.format(url.hostname,
                                                                          app.config.get("OTS_MARTI_HTTPS_PORT"),
                                                                          file_hash), 200

        else:
            return jsonify({'success': False, 'error': 'Something went wrong'}), 400


@marti_blueprint.route('/Marti/api/sync/metadata/<file_hash>/tool', methods=['GET', 'PUT'])
def data_package_metadata(file_hash):
    if request.method == 'PUT':
        try:
            data_package = db.session.execute(db.select(DataPackage).filter_by(hash=file_hash)).scalar_one()
            if data_package:
                data_package.keywords = bleach.clean(request.data.decode("utf-8"))
                db.session.add(data_package)
                db.session.commit()
                return '', 200
            else:
                return '', 404
        except BaseException as e:
            logger.error("Data package PUT failed: {}".format(e))
            logger.error(traceback.format_exc())
            return {'error': str(e)}, 500
    elif request.method == 'GET':
        data_package = db.session.execute(db.select(DataPackage).filter_by(hash=file_hash)).scalar_one()
        return send_from_directory(app.config.get("UPLOAD_FOLDER"), data_package.hash + ".zip",
                                   download_name=data_package.filename)


@marti_blueprint.route('/Marti/sync/missionquery')
def data_package_query():
    try:
        data_package = db.session.execute(db.select(DataPackage).filter_by(hash=request.args.get('hash'))).scalar_one()
        if data_package:

            url = urlparse(request.url_root)
            return 'https://{}:{}/Marti/api/sync/metadata/{}/tool'.format(url.hostname,
                                                                          app.config.get("OTS_MARTI_HTTPS_PORT"),
                                                                          request.args.get('hash')), 200
        else:
            return {'error': '404'}, 404, {'Content-Type': 'application/json'}
    except sqlalchemy.exc.NoResultFound as e:
        return {'error': '404'}, 404, {'Content-Type': 'application/json'}


@marti_blueprint.route('/Marti/sync/search', methods=['GET'])
def data_package_search():
    data_packages = db.session.execute(db.select(DataPackage)).scalars()
    res = {'resultCount': 0, 'results': []}
    for dp in data_packages:
        submission_user = "anonymous"
        if dp.user:
            submission_user = dp.user.username
        res['results'].append(
            {'UID': dp.hash, 'Name': dp.filename, 'Hash': dp.hash, 'CreatorUid': dp.creator_uid,
             "SubmissionDateTime": dp.submission_time.strftime('%Y-%m-%dT%H:%M:%S.000Z'), "EXPIRATION": "-1",
             "Keywords": ["missionpackage"],
             "MIMEType": dp.mime_type, "Size": "{}".format(dp.size), "SubmissionUser": submission_user,
             "PrimaryKey": "{}".format(dp.id),
             "Tool": dp.tool if dp.tool else "public"
             })
        res['resultCount'] += 1

    return json.dumps(res), 200, {'Content-Type': 'application/json'}


@marti_blueprint.route('/Marti/sync/content', methods=['GET'])
def download_data_package():
    file_hash = request.args.get('hash')
    data_package = db.session.execute(db.select(DataPackage).filter_by(hash=file_hash)).scalar_one()

    return send_from_directory(app.config.get("UPLOAD_FOLDER"), file_hash + ".zip", download_name=data_package.filename)


@marti_blueprint.route('/Marti/vcm', methods=['GET', 'POST'])
def video():
    if request.method == 'POST':
        soup = BeautifulSoup(request.data, 'xml')
        video_connections = soup.find('videoConnections')

        path = video_connections.find('path').text
        if path.startswith("/"):
            path = path[1:]

        if video_connections:
            v = VideoStream()
            v.protocol = video_connections.find('protocol').text
            v.alias = video_connections.find('alias').text
            v.uid = video_connections.find('uid').text
            v.port = video_connections.find('port').text
            v.rover_port = video_connections.find('roverPort').text
            v.ignore_embedded_klv = (video_connections.find('ignoreEmbeddedKLV').text.lower() == 'true')
            v.preferred_mac_address = video_connections.find('preferredMacAddress').text
            v.preferred_interface_address = video_connections.find('preferredInterfaceAddress').text
            v.path = path
            v.buffer_time = video_connections.find('buffer').text
            v.network_timeout = video_connections.find('timeout').text
            v.rtsp_reliable = video_connections.find('rtspReliable').text
            path_config = MediaMTXPathConfig(None).serialize()
            path_config['sourceOnDemand'] = False
            v.mediamtx_settings = json.dumps(path_config)

            # Discard username and password for security
            feed = soup.find('feed')
            address = feed.find('address').text
            feed.find('address').string.replace_with(address.split("@")[-1])

            v.xml = str(feed)

            with app.app_context():
                try:
                    db.session.add(v)
                    db.session.commit()
                    logger.debug("Inserted Video")
                except sqlalchemy.exc.IntegrityError as e:
                    db.session.rollback()
                    v = db.session.execute(db.select(VideoStream).filter_by(path=v.path)).scalar_one()
                    v.protocol = video_connections.find('protocol').text
                    v.alias = video_connections.find('alias').text
                    v.uid = video_connections.find('uid').text
                    v.port = video_connections.find('port').text
                    v.rover_port = video_connections.find('roverPort').text
                    v.ignore_embedded_klv = (video_connections.find('ignoreEmbeddedKLV').text.lower() == 'true')
                    v.preferred_mac_address = video_connections.find('preferredMacAddress').text
                    v.preferred_interface_address = video_connections.find('preferredInterfaceAddress').text
                    v.path = video_connections.find('path').text
                    v.buffer_time = video_connections.find('buffer').text
                    v.network_timeout = video_connections.find('timeout').text
                    v.rtsp_reliable = video_connections.find('rtspReliable').text
                    feed = soup.find('feed')
                    address = feed.find('address').text
                    feed.find('address').replace_with(address.split("@")[-1])

                    v.xml = str(feed)

                    db.session.commit()
                    logger.debug("Updated video")

        return '', 200

    elif request.method == 'GET':
        try:
            with app.app_context():
                videos = db.session.execute(db.select(VideoStream)).scalars()
                videoconnections = Element('videoConnections')

                for video in videos:
                    # Make sure videos have the correct address based off of the Flask request and not 127.0.0.1
                    # This also forces all streams to bounce through MediaMTX
                    feed = BeautifulSoup(video.xml, 'xml')

                    url = urlparse(request.url_root).hostname
                    path = feed.find('path').text
                    if not path.startswith("/"):
                        path = "/" + path

                    if 'iTAK' in request.user_agent.string:
                        url = feed.find('protocol').text + "://" + url + ":" + feed.find("port").text + path

                    if feed.find('address'):
                        feed.find('address').string.replace_with(url)
                    else:
                        address = feed.new_tag('address')
                        address.string = url
                        feed.feed.append(address)
                    videoconnections.append(fromstring(str(feed)))

            return tostring(videoconnections), 200
        except BaseException as e:
            logger.error(traceback.format_exc())
            return '', 500



from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, BooleanField


class MediaMTSGlobalConfig(FlaskForm):
    logLevel = StringField()
    #logDestinations
    logFile = StringField()
    readTimeout = StringField()
    writeTimeout = StringField()
    writeQueueSize = IntegerField()
    udpMaxPayloadSize = IntegerField()
    externalAuthenticationURL = StringField()
    api = BooleanField()
    apiAddress = StringField()
    metrics = BooleanField()
    metricsAddress = StringField()
    pprof = BooleanField()
    pprofAddress = StringField()
    runOnConnect = StringField()
    runOnConnectRestart = BooleanField()
    runOnDisconnect = StringField()
    rtsp = BooleanField()
    #protocols
    encryption = StringField()
    rtspAddress = StringField()
    rtspsAddress = StringField()
    rtpAddress = StringField()
    rtcpAddress = StringField()
    multicastIPRange = StringField()
    multicastRTPPort = IntegerField()
    multicastRTCPPort = IntegerField()
    serverKey = StringField()
    serverCert = StringField()
    #authMethods
    rtmp = BooleanField()
    rtmpAddress = StringField()
    rtmpEncryption = StringField()
    rtmpsAddress = StringField()
    rtmpServerKey = StringField()
    rtmpServerCert = StringField()
    hls = BooleanField()
    hlsAddress = StringField()
    hlsEncryption = BooleanField()
    hlsServerKey = StringField()
    hlsServerCert = StringField()
    hlsAlwaysRemux = BooleanField()
    hlsVariant = StringField()
    hlsSegmentCount = IntegerField()
    hlsSegmentDuration = StringField()
    hlsPartDuration = StringField()
    hlsSegmentMaxSize = StringField()
    hlsAllowOrigin = StringField()
    #hlsTrustedProxies
    hlsDirectory = StringField()
    webrtc = BooleanField()
    webrtcAddress = StringField()
    webrtcEncryption = BooleanField()
    webrtcServerKey = StringField()
    webrtcServerCert = StringField()
    webrtcAllowOrigin = StringField()
    #webrtcTrustedProxies
    webrtcLocalUDPAddress = StringField()
    webrtcLocalTCPAddress = StringField()
    webrtcIPsFromInterfaces = BooleanField()
    #webrtcIPsFromInterfacesList
    #webrtcAdditionalHosts
    #webrtcICEServers2
    srt = BooleanField()
    srtAddress = StringField()

    def serialize(self):
        return_value = {}
        for field in self._fields:
            if field != 'csrf_token':
                return_value[field] = self._fields[field].data
        return return_value
